# Authors: Isak Samsten
# License: BSD 3 clause

import sys
from abc import ABCMeta, abstractmethod
from numbers import Integral, Real

import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.utils import check_random_state, compute_sample_weight
from sklearn.utils._param_validation import Interval
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import _check_sample_weight, check_is_fitted

from ..base import BaseEstimator
from ..utils.validation import _check_ts_array, _num_timesteps


# noqa: PR01
class BaseTree(BaseEstimator, metaclass=ABCMeta):
    """Base class for tree based estimators."""

    _parameter_constraints: dict = {
        "max_depth": [
            Interval(Integral, 1, sys.getrecursionlimit(), closed="left"),
            None,
        ],
        "min_samples_split": [Interval(Integral, 2, None, closed="left")],
        "min_samples_leaf": [Interval(Integral, 1, None, closed="left")],
        "min_impurity_decrease": [Interval(Real, 0.0, None, closed="left")],
    }

    def __init__(
        self,
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
    ):
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_impurity_decrease = min_impurity_decrease

    def _validate_not_check_input(self, x, reset=True):
        x = self._validate_force_n_dims(x)
        if reset:
            self.n_timesteps_in_, self.n_dims_in_ = _num_timesteps(x)
            self.n_features_in_ = self.n_timesteps_in_ * self.n_dims_in_

        return _check_ts_array(x)

    @abstractmethod
    def _fit(self, X, y, sample_weight, max_depth, random_state):
        pass

    def decision_path(self, x, check_input=True):
        """
        Compute the decision path of the tree.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or\
 (n_samples, n_dims, n_timestep)
            The input samples.
        check_input : bool, optional
            Bypass array validation. Only set to True if you are sure your data
            is valid.

        Returns
        -------
        sparse matrix of shape (n_samples, n_nodes)
            An indicator array where each nonzero values indicate that the sample
            traverses a node.
        """
        check_is_fitted(self, attributes="tree_")
        if check_input:
            x = self._validate_data(
                x, dtype=float, ensure_ts_array=True, allow_3d=True, reset=False
            )
        else:
            x = self._validate_not_check_input(x, reset=False)

        return self.tree_.decision_path(x)

    def apply(self, x, check_input=True):
        """
        Return the index of the leaf that each sample is predicted by.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or\
 (n_samples, n_dims, n_timestep)
            The input samples.
        check_input : bool, optional
            Bypass array validation. Only set to True if you are sure your data
            is valid.

        Returns
        -------
        ndarray of shape (n_samples, )
            For every sample, return the index of the leaf that the sample
            ends up in. The index is in the range [0; node_count].

        Examples
        --------
        Get the leaf probability distribution of a prediction:

        >>> from wildboar.datasets import load_gun_point
        >>> from wildboar.tree import ShapeletTreeClassifier
        >>> X, y = load_gun_point()
        >>> tree = ShapeletTreeClassifier()
        >>> tree.fit(X, y)
        >>> leaves = tree.apply(X)
        >>> tree.tree_.value.take(leaves, axis=0)
        array([[0., 1.],
               [0., 1.],
               [1., 0.]])

        This is equvivalent to using `tree.predict_proba`.
        """
        check_is_fitted(self, attributes="tree_")
        if check_input:
            x = self._validate_data(
                x, dtype=float, ensure_ts_array=True, allow_3d=True, reset=False
            )
        else:
            x = self._validate_not_check_input(x, reset=False)

        return self.tree_.apply(x)

    def _more_tags(self):
        return {"X_types": ["2darray", "3darray"]}


class BaseTreeRegressor(RegressorMixin, BaseTree, metaclass=ABCMeta):
    def fit(self, x, y, sample_weight=None, check_input=True):
        """
        Fit the estimator.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps)
            The training time series.
        y : array-like of shape (n_samples,)
            Target values as floating point values.
        sample_weight : array-like of shape (n_samples,), optional
            If `None`, then samples are equally weighted. Splits that would create child
            nodes with net zero or negative weight are ignored while searching for a
            split in each node. Splits are also ignored if they would result in any
            single class carrying a negative weight in either child node.
        check_input : bool, optional
            Allow to bypass several input checks.

        Returns
        -------
        self
            This object.

        """
        self._validate_params()
        if check_input:
            x, y = self._validate_data(
                x, y, allow_3d=True, ensure_ts_array=True, dtype=float, y_numeric=True
            )
        else:
            x = self._validate_not_check_input(x)

        random_state = check_random_state(self.random_state)
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, x, dtype=float)

        self._fit(
            x,
            y,
            sample_weight,
            min(self.max_depth or sys.getrecursionlimit(), sys.getrecursionlimit()),
            random_state,
        )
        return self

    def predict(self, x, check_input=True):
        """
        Predict the value of x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps)
            The input time series.
        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you
            know what you do.

        Returns
        -------
        ndarray of shape (n_samples,)
            The predicted classes.

        """
        check_is_fitted(self, ["tree_"])
        if check_input:
            x = self._validate_data(
                x, allow_3d=True, ensure_ts_array=True, dtype=float, reset=False
            )
        else:
            x = self._validate_not_check_input(x, reset=False)

        return self.tree_.predict(x).reshape(-1)


class BaseTreeClassifier(ClassifierMixin, BaseTree, metaclass=ABCMeta):
    """Mixin for classification trees."""

    def fit(self, x, y, sample_weight=None, check_input=True):
        """
        Fit a classification tree.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps)
            The training time series.
        y : array-like of shape (n_samples,)
            The target values.
        sample_weight : array-like of shape (n_samples,), optional
            If `None`, then samples are equally weighted. Splits that would create child
            nodes with net zero or negative weight are ignored while searching for a
            split in each node. Splits are also ignored if they would result in any
            single class carrying a negative weight in either child node.
        check_input : bool, optional
            Allow to bypass several input checks.

        Returns
        -------
        self
            This instance.

        """
        self._validate_params()
        if check_input:
            x, y = self._validate_data(
                x, y, allow_3d=True, ensure_ts_array=True, dtype=float
            )
            check_classification_targets(y)
        else:
            x = self._validate_not_check_input(x)

        if hasattr(self, "class_weight") and self.class_weight is not None:
            class_weight = compute_sample_weight(self.class_weight, y)
        else:
            class_weight = None

        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        random_state = check_random_state(
            self.random_state if hasattr(self, "random_state") else None
        )

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, x, dtype=float)

        if class_weight is not None:
            if sample_weight is not None:
                sample_weight = sample_weight * class_weight
            else:
                sample_weight = class_weight

        self._fit(
            x,
            y,
            sample_weight,
            min(self.max_depth or sys.getrecursionlimit(), sys.getrecursionlimit()),
            random_state,
        )
        return self

    def predict(self, x, check_input=True):
        """
        Predict the regression of the input samples x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps)
            The input time series.
        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you
            know what you do.

        Returns
        -------
        ndarray of shape (n_samples,)
            The predicted classes.

        """
        proba = self.predict_proba(x, check_input=check_input)
        return self.classes_.take(np.argmax(proba, axis=1), axis=0)

    def predict_proba(self, x, check_input=True):
        """
        Predict class probabilities of the input samples X.

        The predicted class probability is the fraction of samples of the same
        class in a leaf.

        Parameters
        ----------
        x :  array-like of shape (n_samples, n_timesteps)
            The input time series.
        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you
            know what you do.

        Returns
        -------
        ndarray of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the classes
            corresponds to that in the attribute `classes_`.

        """
        check_is_fitted(self, ["tree_"])
        if check_input:
            x = self._validate_data(
                x, ensure_ts_array=True, allow_3d=True, dtype=float, reset=False
            )
        else:
            x = self._validate_not_check_input(x, reset=False)

        return self.tree_.predict(x)



from numbers import Integral, Real

from sklearn.manifold import MDS as Sklearn_MDS  # noqa: N811
from sklearn.utils._param_validation import Interval, StrOptions

from ..base import BaseEstimator
from ..distance._distance import _METRICS, pairwise_distance


class MDS(BaseEstimator):
    """
    Multidimensional scaling.

    Parameters
    ----------
    n_components : int, optional
        Number of dimensions in which to immerse the dissimilarities.
    metric : bool, optional
        If `True`, perform metric MDS; otherwise, perform nonmetric MDS.
        When `False` (i.e. non-metric MDS), dissimilarities with 0 are considered as
        missing values.
    n_init : int, optional
        Number of times the SMACOF algorithm will be run with different
        initializations. The final results will be the best output of the runs,
        determined by the run with the smallest final stress.
    max_iter : int, optional
        Maximum number of iterations of the SMACOF algorithm for a single run.
    verbose : int, optional
        Level of verbosity.
    eps : float, optional
        Relative tolerance with respect to stress at which to declare
        convergence. The value of `eps` should be tuned separately depending
        on whether or not `normalized_stress` is being used.
    n_jobs : int, optional
        The number of jobs to use for the computation. If multiple
        initializations are used (``n_init``), each run of the algorithm is
        computed in parallel.
    random_state : int, RandomState instance or None, optional
        Determines the random number generator used to initialize the centers.
        Pass an int for reproducible results across multiple function calls.
    dissimilarity : str, optional
        The dissimilarity measure.

        See `_METRICS.keys()` for a list of supported metrics.
    dissimilarity_params : dict, optional
        Parameters to the dissimilarity measue.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    normalized_stress : bool or "auto", optional
        Whether use and return normed stress value (Stress-1) instead of raw
        stress calculated by default. Only supported in non-metric MDS.

    Notes
    -----
    This implementation is a convenience wrapper around
    :class:`sklearn.manifold.MDS` to when using Wildboar metrics.
    """

    _parameter_constraints: dict = {
        "n_components": [Interval(Integral, 1, None, closed="left")],
        "metric": ["boolean"],
        "n_init": [Interval(Integral, 1, None, closed="left")],
        "max_iter": [Interval(Integral, 1, None, closed="left")],
        "verbose": ["verbose"],
        "eps": [Interval(Real, 0.0, None, closed="left")],
        "n_jobs": [None, Integral],
        "random_state": ["random_state"],
        "dissimilarity": [StrOptions(_METRICS.keys())],
        "dissimilarity_params": [None, dict],
        "normalized_stress": [
            "boolean",
            StrOptions({"auto"}),
        ],
    }

    def __init__(
        self,
        n_components=2,
        *,
        metric=True,
        n_init=4,
        max_iter=300,
        verbose=0,
        eps=1e-3,
        n_jobs=None,
        random_state=None,
        dissimilarity="euclidean",
        dissimilarity_params=None,
        normalized_stress="warn",
    ):
        self.n_components = n_components
        self.metric = metric
        self.n_init = n_init
        self.max_iter = max_iter
        self.verbose = verbose
        self.eps = eps
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.normalized_stress = normalized_stress
        self.dissimilarity = dissimilarity
        self.dissimilarity_params = dissimilarity_params

    def fit(self, x, y=None):
        x = self._validate_data(x, allow_3d=True)
        self.fit_transform(x)
        return self

    def fit_transform(self, x, y=None):
        self._validate_params()
        x = self._validate_data(x, allow_3d=True, reset=False)
        self.mds_ = Sklearn_MDS(
            n_components=self.n_components,
            metric=self.metric,
            n_init=self.n_init,
            max_iter=self.max_iter,
            verbose=self.verbose,
            eps=self.eps,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            dissimilarity="precomputed",
            normalized_stress=self.normalized_stress,
        )
        return self.mds_.fit_transform(
            pairwise_distance(
                x,
                dim="mean",
                metric=self.dissimilarity,
                n_jobs=self.n_jobs,
                metric_params=self.dissimilarity_params,
            )
        )



# Authors: Isak Samsten
# License: BSD 3 clause
from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.cluster import KMeans as Sklearn_KMeans
from sklearn.metrics.pairwise import pairwise_distances as sklearn_pairwise_distances
from sklearn.neighbors import KNeighborsClassifier as Sklearn_KNeighborsClassifier
from sklearn.utils._param_validation import StrOptions
from sklearn.utils.validation import check_is_fitted, check_random_state

from ...base import BaseEstimator, CounterfactualMixin, ExplainerMixin
from ...distance._distance import pairwise_distance
from ...distance._neighbors import KMeans, KMedoids


class _Explainer(metaclass=ABCMeta):
    def __init__(self, n_neighbors, n_cluster, metric, metric_params, random_state):
        self.n_clusters = n_cluster
        self.metric = metric
        self.metric_params = metric_params
        self.random_state = random_state
        self.n_neighbors = n_neighbors

    @abstractmethod
    def _cluster(self, x):
        pass

    @abstractmethod
    def _pairwise_distance(self, x, y):
        pass

    def _assign(self, x, y, classes):
        if not hasattr(self, "cluster_centers_"):
            raise ValueError("_cluster must be called before _assign")

        n_classes = len(classes)
        self.majority_centers_ = {}

        # Find top n_neighbors labels from the traning samples
        # for the cluster centers
        cluster_labels = np.take(
            y,
            np.argpartition(
                self._pairwise_distance(self.cluster_centers_, x),
                self.n_neighbors,
                axis=1,
            )[:, : self.n_neighbors],
        )

        for cls in classes:
            cls_idx = np.nonzero(classes == cls)[0][0]
            # Store the cluster centers that would be predicted as cls
            self.majority_centers_[cls] = self.cluster_centers_[
                (cluster_labels == cls_idx).sum(axis=1) >= self.n_neighbors // 2 + 1
            ]

    def _find_closest_center(self, x, y):
        if not hasattr(self, "majority_centers_"):
            raise ValueError("_assign must be called before _find_closest_center")

        centers = self.majority_centers_[y]
        if centers.shape[0] == 0:
            return None

        return centers[
            self._pairwise_distance(x, centers).argmin(axis=1),
            :,
        ]


class _SklearnExplainer(_Explainer):
    def _pairwise_distance(self, x, y):
        metric_params = self.metric_params if self.metric_params is not None else {}
        return sklearn_pairwise_distances(x, y, metric=self.metric, **metric_params)


class _SklearnKMeansExplainer(_SklearnExplainer):
    def _cluster(self, x):
        kmeans = Sklearn_KMeans(
            n_clusters=self.n_clusters, n_init=10, random_state=self.random_state
        ).fit(x)

        self.cluster_centers_ = kmeans.cluster_centers_
        self.labels_ = kmeans.labels_


class _SklearnKMedoidsExplainer(_SklearnExplainer):
    def _cluster(self, x):
        metric_params = self.metric_params if self.metric_params is not None else {}
        kmedoids = KMedoids(
            n_clusters=self.n_clusters,
            metric="precomputed",
            metric_params=self.metric_params,
            random_state=self.random_state,
        ).fit(sklearn_pairwise_distances(x, metric=self.metric, **metric_params))

        self.cluster_centers_ = x[kmedoids.medoid_indices_, :]
        self.labels_ = kmedoids.labels_


class _WildboarExplainer(_Explainer):
    def _pairwise_distance(self, x, y):
        return pairwise_distance(
            x,
            y,
            metric=self.metric,
            metric_params=self.metric_params,
            dim="mean",
        )


class _WildboarKMeansExplainer(_WildboarExplainer):
    def _cluster(self, x):
        metric = self.metric
        if self.metric_params is not None:
            r = self.metric_params.get("r", 1.0)
            g = self.metric_params.get("g")
        else:
            r = 1.0
            g = None

        if self.metric == "wdtw":
            metric = "dtw"

        kmeans = KMeans(
            n_clusters=self.n_clusters,
            metric=metric,
            r=r,
            g=g,
            random_state=self.random_state,
        ).fit(x)
        self.cluster_centers_ = kmeans.cluster_centers_
        self.labels_ = kmeans.labels_


class _WildboarKMedoidsExplainer(_WildboarExplainer):
    def _cluster(self, x):
        kmedoids = KMedoids(
            n_clusters=self.n_clusters,
            metric=self.metric,
            metric_params=self.metric_params,
            random_state=self.random_state,
        ).fit(x)
        self.cluster_centers_ = kmedoids.cluster_centers_
        self.labels_ = kmedoids.labels_


class KNeighborsCounterfactual(CounterfactualMixin, ExplainerMixin, BaseEstimator):
    """
    Fit a counterfactual explainer to a k-nearest neighbors classifier.

    Parameters
    ----------
    method : {"auto", "mean", "medoid"}, optional
        The method for generating counterfactuals. If 'auto', counterfactuals
        are generated using k-means if possible and k-medoids otherwise. If
        'mean', counterfactuals are always generated using k-means, which fails
        if the estimator is fitted with a metric other than 'euclidean', 'dtw'
        or 'wdtw. If 'medoid', counterfactuals are generated using k-medoids.

        .. versionadded:: 1.2
    random_state : int or RandomState, optional
        - If `int`, `random_state` is the seed used by the random number generator.
        - If `RandomState` instance, `random_state` is the random number generator.
        - If `None`, the random number generator is the `RandomState` instance used
          by `np.random`.

    Attributes
    ----------
    explainer_ : dict
        The explainer for each label

    References
    ----------
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.
    """

    _parameter_constraints: dict = {
        "method": [StrOptions({"auto", "mean", "medoid"})],
        "random_state": ["random_state"],
    }

    def __init__(self, method="auto", random_state=None):
        self.random_state = random_state
        self.method = method

    def _validate_estimator(self, estimator, allow_3d=False):
        estimator = super()._validate_estimator(estimator, allow_3d)
        for attr in ["_fit_X", "_y", "n_neighbors", "classes_", "metric"]:
            if not hasattr(estimator, attr):
                raise ValueError(f"estimator must be have the attributes {attr}.")

        return estimator

    def fit(self, estimator, x=None, y=None):
        self._validate_params()
        estimator = self._validate_estimator(estimator)
        x = estimator._fit_X
        y = estimator._y
        self.n_timesteps_in_ = estimator.n_features_in_
        self.n_dims_in_ = 1
        self.n_features_in_ = self.n_timesteps_in_

        n_clusters = x.shape[0] // estimator.n_neighbors

        random_state = check_random_state(self.random_state)

        if isinstance(estimator, Sklearn_KNeighborsClassifier):
            if (
                estimator.metric == "euclidean"
                or (estimator.p == 2 and estimator.metric == "minkowski")
            ) and self.method in ("auto", "mean"):
                Explainer = _SklearnKMeansExplainer
            elif self.method in ("auto", "medoid"):
                Explainer = _SklearnKMedoidsExplainer
            else:
                raise ValueError(
                    (
                        "if method='mean', estimator %r must be fitted with "
                        "metric 'euclidean' or 'minkowski' with 'p=1', got %r"
                    )
                    % (estimator.__class__.__qualname__, estimator.metric)
                )
        elif estimator.metric in ("euclidean", "dtw", "wdtw") and self.method in (
            "auto",
            "mean",
        ):
            Explainer = _WildboarKMeansExplainer
        elif self.method in ("auto", "medoid"):
            Explainer = _WildboarKMedoidsExplainer
        else:
            raise ValueError(
                (
                    "if method='mean', estimator %r must be fitted with metric in "
                    "'euclidean', 'dtw' or 'wdtw', got %r"
                )
                % (estimator.__class__.__qualname__, estimator.metric)
            )

        self.explainer_ = Explainer(
            n_neighbors=estimator.n_neighbors,
            n_cluster=n_clusters,
            metric=estimator.metric,
            metric_params=estimator.metric_params
            if hasattr(estimator, "metric_params")
            else None,
            random_state=random_state,
        )

        self.explainer_._cluster(x)
        self.explainer_._assign(x, y, estimator.classes_)
        return self

    def explain(self, x, y):
        check_is_fitted(self)
        x, y = self._validate_data(x, y, reset=False, dtype=float)
        x_counterfactuals = x.copy()
        labels = np.unique(y)
        for label in labels:
            label_indices = np.where(y == label)[0]

            closest = self.explainer_._find_closest_center(x[label_indices], label)
            if closest is not None:
                x_counterfactuals[label_indices, :] = closest

        return x_counterfactuals

    def _more_tags():
        return {"X_types": []}



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef double scaled_euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    double s_mean,
    double s_std,
    const double *T,
    Py_ssize_t t_length,
    double *X_buffer,
    Py_ssize_t *index,
) noexcept nogil

cdef double euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil

cdef double normalized_euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t *index,
) noexcept nogil

cdef double manhattan_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil

cdef Py_ssize_t manhattan_distance_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) noexcept nogil

cdef double minkowski_distance(
    double p,
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil

cdef double chebyshev_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil

cdef double cosine_similarity(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t *index,
) nogil




# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False
import numpy as np
from libc.math cimport floor
from libc.stdlib cimport labs

from ..utils cimport TSArray

# TODO: Consider a fast code-path for:
#   - dilation = 1
#   - stride = 1
#   - padding = 0
cdef void convolution_1d(
    Py_ssize_t stride,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double bias,
    double *kernel,
    Py_ssize_t k_len,
    const double* x,
    Py_ssize_t x_len,
    double* out,
) noexcept nogil:
    """
    Compute the 1d convolution.

    Parameters
    ----------
    stride : int
        The stride in x.
    dilation : int
        Inflate the kernel by inserting spaces between kernel values.
    padding : int
        Increase the size of the input by padding.
    bias : double
        The bias.
    kernel : double*
        The kernel values.
    k_len : int
        The length of the kernel.
    x : double*
        The input sample.
    x_len : int
        The length of the sample.
    out : double*, output
        The output buffer.

        The invariant is not checked.

    Warnings
    --------
    This code performs no bounds checks. Ensure that the following invariants hold:

    The dilated kernel size must be smaller than the padded sample:::

        (k_len - 1) * dilation - 1 <= x_len + 2 * padding

    The output buffer must be allocated with enough memory:::

        ((x_len + 2 * padding) - (k_len - 1) * dilation + 1) / stride + 1

    """
    # Fast-path for the simple case.
    if padding == 0 and stride == 1 and dilation == 1:
        _convolution_1d_fast(bias, x, x_len, kernel, k_len, out)
        return

    cdef Py_ssize_t input_size = x_len + 2 * padding
    cdef Py_ssize_t kernel_size = (k_len - 1) * dilation + 1
    cdef Py_ssize_t output_size = <Py_ssize_t> floor((input_size - kernel_size) / stride) + 1

    cdef Py_ssize_t j  # the index in the kernel and input array
    cdef Py_ssize_t i  # the index of the output array
    cdef Py_ssize_t padding_offset
    cdef Py_ssize_t input_offset
    cdef Py_ssize_t kernel_offset
    cdef Py_ssize_t convolution_size
    cdef double inner_prod

    for i in range(output_size):
        padding_offset = padding - i * stride

        # this part of the code ensures that the iterators responsible
        # for selecting the correct values in `kernel` and `x` start
        # at the correct location. the main idea is to always start
        # the index at the first non-dilated index *after* padding.
        #
        # example:
        #  k = [1, 1, 2]
        #  d = 2
        #  k_d = [1, 0, 1, 0, 2]
        #  x = [1, 2, 3, 4, 5, 6]
        #  p = 2
        #  x_p = [0, 0, 1, 2, 3, 4, 5, 6, 0, 0]
        #
        #  first convoluton (i = 0):
        #
        #          / start the kernel index here, since p - 1 * s % d == 0,
        #          | that is: 2 % 2 == 0
        #          |
        # k_p  1 0 1 0 2
        #      | | | | |
        # x_p  0 0 1 2 3 4 5 6 0 0 (result: 7)
        #      |   |
        #      |   \ start the input index here, which is given by
        #      |     kernel_offset - padding_offset, that is 2 - 2 = 0
        #      |
        #      \ padding_offset is 2, so we are at the first padded value,
        #        but we can ignore this part since the pad is all zeros
        #        so we move head of the kernel iterator to first non-dilated
        #        value, which is located at imaginary kernel index 2, which also
        #        happens to be at the start of the input.
        #
        #  second convolution
        #
        #
        #            / since 1 % 2 != 0, we move the start of the kernel to the
        #            | the first non-dilated value, which is given by
        #            | 1 + 2 - (1 % 2) = 2
        #            |
        # k_p    1 0 1 0 2
        #        | | | | |
        # x_p  0 0 1 2 3 4 5 6 0 0 (result: 10)
        #        | |
        #        | \ we move the input iterator the the index where the
        #        |   first non dilated value is, kernel_offset (2) - padding_offset (1)
        #        \
        #         padding_offset = 1, so we are at the second padded value.
        #
        #  third convolution
        #
        #          / padding_offset = 0, so we should start the kernel
        #          | with the first value (which by definition is non-dilated).
        #          |
        # k_p      1 0 1 0 2
        #          | | | | |
        # x_p  0 0 1 2 3 4 5 6 0 0 (result: 14)
        #          |
        #          \ we should start x at the current (strided) index.
        #
        # we continue iterating, until we have visited all start locations
        # but stopping the convolution at the end of the input array.
        if padding_offset > 0:
            if padding_offset % dilation == 0:
                kernel_offset = padding_offset
            else:
                kernel_offset = padding_offset + dilation - (padding_offset % dilation)
            input_offset = kernel_offset - padding_offset
        else:
            kernel_offset = 0
            input_offset = labs(padding_offset)

        # the iteration should be performed up until but not including the
        # padding. so, the last value we should convolve over is either the
        # last value of the input, or a value before that located where the
        # kernel ends. the kernel size here takes into account dilation. we
        # express this as a length from input_offset until the end.
        convolution_size = (
            min(x_len, input_offset + kernel_size - max(0, padding_offset))
            - input_offset
        )
        inner_prod = bias
        for j from 0 <= j < convolution_size by dilation:
            # we adjust the index (j) to account for the input_offset
            # and the kernel_offset. the kernel index is in the range
            # of the dilated kernel so we take the mod to scale it back
            # to the original, non-dilated, array.
            inner_prod += x[input_offset + j] * kernel[((j + kernel_offset) // dilation)]

        out[i] = inner_prod


cdef void _convolution_1d_fast(
    double bias,
    const double *x,
    Py_ssize_t x_len,
    double *kernel,
    Py_ssize_t k_len,
    double *out
) noexcept nogil:
    cdef Py_ssize_t output_size = x_len - k_len + 1
    cdef Py_ssize_t i, j
    cdef double inner_prod
    for i in range(x_len - k_len + 1):
        inner_prod = bias
        for j in range(k_len):
            inner_prod += x[i + j] * kernel[j]
        out[i] = inner_prod


def conv1d(
    TSArray X,
    double[::1] kernel,
    double bias,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    Py_ssize_t stride,
    double[:, ::1] out,
):
    cdef Py_ssize_t i
    with nogil:
        for i in range(X.shape[0]):
            convolution_1d(
                stride,
                dilation,
                padding,
                bias,
                &kernel[0],
                kernel.shape[0],
                &X[i, 0, 0],
                X.shape[2],
                &out[i, 0],
            )



import numbers

from sklearn.utils._param_validation import Interval, StrOptions

from ._base import BaseAttributeTransform
from ._chydra import HydraAttributeGenerator, NormalKernelSampler

_SAMPLING_METHOD = {
    "normal": NormalKernelSampler,
}


class HydraMixin:
    _parameter_constraints: dict = {
        "n_kernels": [Interval(numbers.Integral, 1, None, closed="left")],
        "kernel_size": [Interval(numbers.Integral, 2, None, closed="left")],
        "n_groups": [Interval(numbers.Integral, 1, None, closed="left")],
        "sampling": [StrOptions({"normal"})],
        "sampling_params": [dict, None],
    }

    def _get_generator(self, x, y):
        sampling_params = {} if self.sampling_params is None else self.sampling_params
        return HydraAttributeGenerator(
            self.n_groups,
            self.n_kernels,
            self.kernel_size,
            _SAMPLING_METHOD[self.sampling](**sampling_params),
        )


class HydraTransform(HydraMixin, BaseAttributeTransform):
    """
    A Dictionary based method using convolutional kernels.

    Parameters
    ----------
    n_groups : int, optional
        The number of groups of kernels.
    n_kernels : int, optional
        The number of kernels per group.
    kernel_size : int, optional
        The size of the kernel.
    sampling : {"normal"}, optional
        The strategy for sampling kernels. By default kernel weights
        are sampled from a normal distribution with zero mean and unit
        standard deviation.
    sampling_params : dict, optional
        Parameters to the sampling approach. The "normal" sampler
        accepts two parameters: `mean` and `scale`.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means using
        a single core and a value of `-1` means using all cores. Positive
        integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random sampling of kernels.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding

    See Also
    --------
    HydraClassifier : A classifier using hydra transform.

    Notes
    -----
    The implementation does not implement the first order descrete differences
    described by Dempster et. al. (2023). If this is desired, one can use
    native scikit-learn functionalities and the
    :class:`~wildboar.transform.DiffTransform`:

    >>> from sklearn.pipeline import make_pipeline, make_union
    >>> from wildboar.transform import DiffTransform, HydraTransform
    >>> dempster_hydra = make_union(
    ...     HydraTransform(n_groups=32),
    ...     make_pipeline(
    ...         DiffTransform(),
    ...         HydraTransform(n_groups=32)
    ...     )
    ... )

    References
    ----------
    Dempster, A., Schmidt, D. F., & Webb, G. I. (2023).
        Hydra: competing convolutional kernels for fast and accurate
        time series classification. Data Mining and Knowledge Discovery

    Examples
    --------
    >>> from wildboar.datasets import load_gun_point
    >>> from wildboar.transform import HydraTransform
    >>> X, y = load_gun_point()
    >>> t = HydraTransform(n_groups=8, n_kernels=4, random_state=1)
    >>> t.fit_transform(X)
    """

    _parameter_constraints: dict = {
        **HydraMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        *,
        n_groups=64,
        n_kernels=8,
        kernel_size=9,
        sampling="normal",
        sampling_params=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(n_jobs=n_jobs, random_state=random_state)
        self.n_groups = n_groups
        self.n_kernels = n_kernels
        self.kernel_size = kernel_size
        self.sampling = sampling
        self.sampling_params = sampling_params



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef void fft(complex *x, Py_ssize_t n, double fct) noexcept nogil
cdef void ifft(complex *x, Py_ssize_t n, double fct) noexcept nogil

cdef void rfft(double *x, Py_ssize_t n, double fct) noexcept nogil
cdef void irfft(double *x, Py_ssize_t n, double fct) nogil


"""Ensemble methods for classification, regression and outlier detection."""

# Authors: Isak Samsten
# License: BSD 3 clause

from ._elastic import ElasticEnsembleClassifier
from ._ensemble import (
    BaggingClassifier,
    BaggingRegressor,
    BaseBagging,
    ExtraShapeletTreesClassifier,
    ExtraShapeletTreesRegressor,
    IntervalForestClassifier,
    IntervalForestRegressor,
    IsolationShapeletForest,
    PivotForestClassifier,
    ProximityForestClassifier,
    RocketForestClassifier,
    RocketForestRegressor,
    ShapeletForestClassifier,
    ShapeletForestEmbedding,
    ShapeletForestRegressor,
)

__all__ = [
    "BaseBagging",
    "BaggingClassifier",
    "BaggingRegressor",
    "ShapeletForestClassifier",
    "ExtraShapeletTreesClassifier",
    "ShapeletForestRegressor",
    "ExtraShapeletTreesRegressor",
    "IsolationShapeletForest",
    "ShapeletForestEmbedding",
    "RocketForestClassifier",
    "RocketForestRegressor",
    "IntervalForestClassifier",
    "IntervalForestRegressor",
    "PivotForestClassifier",
    "ProximityForestClassifier",
    "ElasticEnsembleClassifier",
]



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

cimport scipy.linalg.cython_blas as blas
cimport scipy.linalg.cython_lapack as lapack
from libc.math cimport INFINITY, M_PI, acos, cos, fabs, pow, sqrt, isinf
from libc.stdlib cimport free, malloc

from ..utils cimport TSArray
from ..utils._misc cimport realloc_array
from ..utils._stats cimport IncStats, inc_stats_add, inc_stats_init, inc_stats_variance
from ._cdistance cimport (
    EPSILON,
    Metric,
    ScaledSubsequenceMetric,
    Subsequence,
    SubsequenceMetric,
    SubsequenceView,
)


cdef class EuclideanSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return euclidean_distance(s, s_len, x, x_len, INFINITY, return_index)

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return euclidean_distance_matches(
            s, s_len, x, x_len, threshold, distances, indicies,
        )


cdef class NormalizedEuclideanSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return normalized_euclidean_distance(s, s_len, x, x_len, return_index)

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return normalized_euclidean_distance_matches(
            s, s_len, x, x_len, threshold, distances, indicies,
        )


cdef class ScaledEuclideanSubsequenceMetric(ScaledSubsequenceMetric):
    cdef double *X_buffer

    def __init__(self):
        pass

    def __cinit__(self):
        self.X_buffer = NULL

    def __dealloc__(self):
        if self.X_buffer != NULL:
            free(self.X_buffer)
            self.X_buffer = NULL

    def __reduce__(self):
        return self.__class__, ()

    cdef int reset(self, TSArray X) noexcept nogil:
        if self.X_buffer != NULL:
            free(self.X_buffer)

        self.X_buffer = <double*> malloc(sizeof(double) * X.shape[2] * 2)
        if self.X_buffer == NULL:
            return -1

        return 0

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return scaled_euclidean_distance(
            s,
            s_len,
            s_mean,
            s_std,
            x,
            x_len,
            self.X_buffer,
            return_index,
        )

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return scaled_euclidean_distance_matches(
            s,
            s_len,
            s_mean,
            s_std,
            x,
            x_len,
            self.X_buffer,
            threshold,
            distances,
            indicies
        )


cdef class ManhattanSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return manhattan_distance(s, s_len, x, x_len, INFINITY, return_index)

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return manhattan_distance_matches(
            s, s_len, x, x_len, threshold, distances, indicies
        )


cdef class MinkowskiSubsequenceMetric(SubsequenceMetric):
    cdef double p

    def __init__(self, double p=2):
        self.p = p

    def __reduce__(self):
        return self.__class__, (self.p, )

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return minkowski_distance(self.p, s, s_len, x, x_len, INFINITY, return_index)

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return minkowski_distance_matches(
            self.p, s, s_len, x, x_len, threshold, distances, indicies
        )


cdef class ChebyshevSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return chebyshev_distance(s, s_len, x, x_len, INFINITY, return_index)

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        return chebyshev_distance_matches(
            s, s_len, x, x_len, threshold, distances, indicies
        )


cdef class CosineSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        return 1 - cosine_similarity(s, s_len, x, x_len, return_index)

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        cdef Py_ssize_t i
        cdef Py_ssize_t n_matches = cosine_similarity_matches(
            s, s_len, x, x_len, 1 - threshold, distances, indicies
        )

        for i in range(n_matches):
            distances[i] = 1 - distances[i]

        return n_matches


cdef class AngularSubsequenceMetric(SubsequenceMetric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        cdef double cosine = cosine_similarity(s, s_len, x, x_len, return_index)

        # Edge-case where, due to floating point issues, the cosine is larger 1
        # or smaller than -1 we return the min and max value respectivley.
        if cosine > 1:
            return 0
        elif cosine < -1:
            return 1
        else:
            return acos(cosine) / M_PI

    cdef Py_ssize_t _matches(
        self,
        double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        cdef Py_ssize_t i
        threshold = cos(threshold * M_PI) if not isinf(threshold) else -INFINITY
        cdef Py_ssize_t n_matches = cosine_similarity_matches(
            s, s_len, x, x_len, threshold, distances, indicies
        )
        cdef double cosine

        for i in range(n_matches):
            cosine = distances[i]
            if cosine > 1:
                distances[i] = 0
            elif cosine < -1:
                distances[i] = 1
            else:
                distances[i] = acos(cosine) / M_PI

        return n_matches


cdef class EuclideanMetric(Metric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return euclidean_distance(x, x_len, y, y_len, INFINITY, NULL)

    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *min_dist,
    ) noexcept nogil:
        cdef double dist = euclidean_distance(
            x, x_len, y, y_len, min_dist[0] * min_dist[0], NULL
        )
        if dist < min_dist[0]:
            min_dist[0] = dist
            return True
        else:
            return False


cdef class NormalizedEuclideanMetric(Metric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return normalized_euclidean_distance(x, x_len, y, y_len, NULL)


cdef class ManhattanMetric(Metric):
    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return manhattan_distance(x, x_len, y, y_len, INFINITY, NULL)

    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *min_dist,
    ) noexcept nogil:
        cdef double dist = manhattan_distance(
            x, x_len, y, y_len, min_dist[0], NULL
        )
        if dist < min_dist[0]:
            min_dist[0] = dist
            return True
        else:
            return False

cdef class MinkowskiMetric(Metric):
    cdef double p

    def __init__(self, double p=2):
        self.p = p

    def __reduce__(self):
        return self.__class__, (self.p, )

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return minkowski_distance(self.p, x, x_len, y, y_len, INFINITY, NULL)

    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *min_dist,
    ) noexcept nogil:
        cdef double dist = minkowski_distance(
            self.p, x, x_len, y, y_len, pow(min_dist[0], self.p), NULL
        )

        if dist < min_dist[0]:
            min_dist[0] = dist
            return True
        else:
            return False


cdef class ChebyshevMetric(Metric):

    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return chebyshev_distance(x, x_len, y, y_len, INFINITY, NULL)

    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *min_dist,
    ) noexcept nogil:
        cdef double dist = chebyshev_distance(
            x, x_len, y, y_len, min_dist[0], NULL
        )
        if dist < min_dist[0]:
            min_dist[0] = dist
            return True
        else:
            return False


cdef class CosineMetric(Metric):

    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        return 1 - cosine_similarity(x, x_len, y, y_len, NULL)

# https://en.wikipedia.org/wiki/Cosine_similarity#Angular_distance_and_similarity
cdef class AngularMetric(Metric):

    def __init__(self):
        pass

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil:
        cdef double cosine = cosine_similarity(x, x_len, y, y_len, NULL)
        if cosine > 1:
            return 0
        elif cosine < -1:
            return 1
        else:
            return acos(cosine) / M_PI


cdef double scaled_euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    double s_mean,
    double s_std,
    const double *T,
    Py_ssize_t t_length,
    double *X_buffer,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double ex = 0
    cdef double ex2 = 0
    cdef double tmp

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t buffer_pos

    for i in range(t_length):
        current_value = T[i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            tmp = ex2 / s_length - mean * mean
            if tmp > 0:
                std = sqrt(tmp)
            else:
                std = 1.0
            dist = inner_scaled_euclidean_distance(
                s_length,
                s_mean,
                s_std,
                j,
                mean,
                std,
                S,
                X_buffer,
                min_dist,
            )

            if dist < min_dist:
                min_dist = dist
                if index != NULL:
                    index[0] = (i + 1) - s_length

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)


cdef double inner_scaled_euclidean_distance(
    Py_ssize_t length,
    double s_mean,
    double s_std,
    Py_ssize_t j,
    double mean,
    double std,
    const double *X,
    double *X_buffer,
    double min_dist,
) noexcept nogil:
    # Compute the distance between the shapelet (starting at `offset`
    # and ending at `offset + length` normalized with `s_mean` and
    # `s_std` with the shapelet in `X_buffer` starting at `0` and
    # ending at `length` normalized with `mean` and `std`
    cdef double dist = 0
    cdef double x
    cdef Py_ssize_t i

    for i in range(length):
        if dist >= min_dist:
            break
        x = (X[i] - s_mean) / s_std
        x -= (X_buffer[i + j] - mean) / std
        dist += x * x

    return dist


cdef double euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double dist = 0

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist >= min_dist:
                break

            x = T[i + j]
            x -= S[j]
            dist += x * x

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return sqrt(min_dist)

# PTDS (https://stats.stackexchange.com/users/68112/ptds),
#   Definition of normalized Euclidean distance, URL (version: 2021-09-27):
#   https://stats.stackexchange.com/q/498753
cdef double normalized_euclidean_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t *index,
) noexcept nogil:
    cdef IncStats S_stats, T_stats, ST_stats
    cdef double s, t, dist
    cdef double min_dist = INFINITY
    cdef Py_ssize_t i, j
    for i in range(t_length - s_length + 1):
        inc_stats_init(&S_stats)
        inc_stats_init(&T_stats)
        inc_stats_init(&ST_stats)
        for j in range(s_length):
            t = T[i + j]
            s = S[j]
            inc_stats_add(&S_stats, 1.0, s)
            inc_stats_add(&T_stats, 1.0, t)
            inc_stats_add(&ST_stats, 1.0, s - t)

        dist = inc_stats_variance(&S_stats) + inc_stats_variance(&T_stats)
        if dist > 0:
            dist = inc_stats_variance(&ST_stats) / dist

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return sqrt(0.5 * min_dist)


cdef Py_ssize_t normalized_euclidean_distance_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) except -1 nogil:
    cdef double dist = 0
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i, j
    cdef double s, t
    cdef IncStats S_stats, T_stats, ST_stats
    cdef Py_ssize_t n_matches = 0

    for i in range(t_length - s_length + 1):
        inc_stats_init(&S_stats)
        inc_stats_init(&T_stats)
        inc_stats_init(&ST_stats)
        for j in range(s_length):
            t = T[i + j]
            s = S[j]
            inc_stats_add(&S_stats, 1.0, s)
            inc_stats_add(&T_stats, 1.0, t)
            inc_stats_add(&ST_stats, 1.0, s - t)

        dist = inc_stats_variance(&S_stats) + inc_stats_variance(&T_stats)
        if dist > 0:
            dist = sqrt(0.5 * inc_stats_variance(&ST_stats) / dist)

        if dist <= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = dist
            n_matches += 1

    return n_matches


cdef Py_ssize_t euclidean_distance_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) except -1 nogil:
    cdef double dist = 0
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    cdef Py_ssize_t n_matches = 0

    threshold = threshold * threshold
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist > threshold:
                break

            x = T[i + j]
            x -= S[j]
            dist += x * x
        if dist <= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = sqrt(dist)
            n_matches += 1

    return n_matches


cdef Py_ssize_t scaled_euclidean_distance_matches(
   const double *S,
   Py_ssize_t s_length,
   double s_mean,
   double s_std,
   const double *T,
   Py_ssize_t t_length,
   double *X_buffer,
   double threshold,
   double *distances,
   Py_ssize_t *matches,
) noexcept nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0

    cdef double ex = 0
    cdef double ex2 = 0
    cdef double tmp

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t buffer_pos
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t n_matches = 0

    n_matches = 0
    threshold = threshold * threshold
    for i in range(t_length):
        current_value = T[i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            tmp = ex2 / s_length - mean * mean
            if tmp > 0:
                std = sqrt(tmp)
            else:
                std = 1.0
            dist = inner_scaled_euclidean_distance(
                s_length,
                s_mean,
                s_std,
                j,
                mean,
                std,
                S,
                X_buffer,
                threshold,
            )

            if dist <= threshold:
                if matches != NULL:
                    matches[n_matches] = (i + 1) - s_length

                distances[n_matches] = sqrt(dist)
                n_matches += 1

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return n_matches


cdef double manhattan_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double dist = 0

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist >= min_dist:
                break

            dist += fabs(T[i + j] - S[j])

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return min_dist


cdef Py_ssize_t manhattan_distance_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) noexcept nogil:
    cdef double dist = 0
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    cdef Py_ssize_t n_matches = 0

    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist > threshold:
                break

            dist += fabs(T[i + j] - S[j])

        if dist <= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = dist
            n_matches += 1

    return n_matches

cdef double minkowski_distance(
    double p,
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double dist = 0
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist >= min_dist:
                break

            dist += pow(fabs(T[i + j] - S[j]), p)

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return pow(min_dist, 1.0 / p)


cdef Py_ssize_t minkowski_distance_matches(
    double p,
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) noexcept nogil:
    cdef double dist = 0
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    cdef Py_ssize_t n_matches = 0

    threshold = pow(threshold, p)
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist > threshold:
                break

            dist += pow(fabs(T[i + j] - S[j]), p)

        if dist <= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = pow(dist, 1.0 / p)
            n_matches += 1

    return n_matches


cdef double chebyshev_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double min_dist,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double c = 0
    cdef double max_dist

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        max_dist = -INFINITY
        for j in range(s_length):
            if max_dist >= min_dist:
                break

            c = fabs(T[i + j] - S[j])
            max_dist = c if c > max_dist else max_dist

        if max_dist < min_dist:
            min_dist = max_dist
            if index != NULL:
                index[0] = i

    return min_dist


cdef Py_ssize_t chebyshev_distance_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) noexcept nogil:
    cdef double c = 0
    cdef double max_dist
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    cdef Py_ssize_t n_matches = 0

    for i in range(t_length - s_length + 1):
        max_dist = -INFINITY
        for j in range(s_length):
            if max_dist > threshold:
                break

            c = fabs(T[i + j] - S[j])
            max_dist = c if c > max_dist else max_dist

        if max_dist <= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = max_dist
            n_matches += 1

    return n_matches


cdef double cosine_similarity(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t *index,
) noexcept nogil:
    cdef double prod = 0, sim
    cdef double s_norm, t_norm
    cdef double max_sim = -INFINITY

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        prod = 0
        s_norm = 0
        t_norm = 0
        for j in range(s_length):
            prod += T[i + j] * S[j]
            s_norm += pow(S[j], 2)
            t_norm += pow(T[i + j], 2)

        sim = (sqrt(s_norm) * sqrt(t_norm))
        if sim <= EPSILON:
            sim = 0
        else:
            sim = prod / sim

        if sim > max_sim:
            max_sim = sim
            if index != NULL:
                index[0] = i

    return max_sim


cdef Py_ssize_t cosine_similarity_matches(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    double threshold,
    double *distances,
    Py_ssize_t *matches,
) noexcept nogil:
    cdef double prod = 0, sim, s_norm, t_norm
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    cdef Py_ssize_t n_matches = 0

    for i in range(t_length - s_length + 1):
        prod = 0
        s_norm = 0
        t_norm = 0
        for j in range(s_length):
            prod += T[i + j] * S[j]
            s_norm += pow(S[j], 2)
            t_norm += pow(T[i + j], 2)

        sim = prod / (sqrt(s_norm) * sqrt(t_norm))
        if sim >= threshold:
            if matches != NULL:
                matches[n_matches] = i
            distances[n_matches] = sim
            n_matches += 1

    return n_matches



import pytest

from wildboar.datasets import load_datasets


@pytest.mark.parametrize("merge_train_test, length", [(True, 2), (False, 4)])
def test_load_datasets_merge_train_test(merge_train_test, length):
    lengths = [
        len(ret)
        for _, ret in load_datasets(
            "wildboar/ucr-tiny", merge_train_test=merge_train_test
        )
    ]
    assert lengths == [length] * 5



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

from numpy cimport uint32_t

from wildboar.utils cimport TSArray


cdef struct Attribute:
    Py_ssize_t dim       # the dimension of the attribute -1 for undefined
    void* attribute        # the attribute value


cdef class AttributeGenerator:

    # Reset the (allocated) attributes of the
    cdef int reset(self, TSArray X) noexcept nogil

    # Safe to call without calling reset first
    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil

    # Safe to call without calling reset first
    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil

    # Requires `reset`
    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil

    # Initialize a persisent attribute from a transient attribute
    #
    # NOTE: We permit moving of ownership of `transient.attribute`.
    #       If is unsafe to use `transient.attribute` after `init_persistent`
    #       has been called.
    # Requires: `reset`
    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil

    # Calculate the attribute value for sample using the transient attribute
    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil

    # Calculate the attribute value for sample using the persistent attribute
    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil

    # Calculate the attribute value for all samples using the transient featuer
    cdef void transient_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil

    # Calculate the attribute value for all samples using the persistent attribute
    cdef void persistent_values(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        double* values
    ) noexcept nogil

    cdef object persistent_to_object(self, Attribute *attribute)

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute)



"""Linear methods for both classification and regression."""

# Authors: Isak Samsten
# License: BSD 3 clause

from ._hydra import HydraClassifier
from ._rocket import RocketClassifier, RocketRegressor
from ._shapelet import (
    CastorClassifier,
    DilatedShapeletClassifier,
    RandomShapeletClassifier,
    RandomShapeletRegressor,
)

__all__ = [
    "HydraClassifier",
    "RocketClassifier",
    "RocketRegressor",
    "RandomShapeletClassifier",
    "RandomShapeletRegressor",
    "DilatedShapeletClassifier",
    "CastorClassifier",
]



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

import numpy as np

from libc.math cimport INFINITY, floor, log2, pow, sqrt, fabs
from libc.stdlib cimport free, malloc, labs
from libc.string cimport memset

from numpy cimport uint32_t

from ._attr_gen cimport Attribute, AttributeGenerator
from ._cconv cimport convolution_1d

from ..utils cimport TSArray
from ..utils._rand cimport rand_normal

# Hydra group
cdef struct Hydra:
    # size: kernel_size * n_kernels
    double *kernels

cdef class KernelSampler:

    cdef void sample(
        self,
        double *data,
        Py_ssize_t length,
        double *mean,
        double *sum_abs,
        uint32_t *seed
    ) noexcept nogil:
        pass


cdef class NormalKernelSampler(KernelSampler):
    cdef double mean
    cdef double scale

    def __init__(self, mean=0.0, scale=1.0):
        self.mean = mean
        self.scale = scale

    cdef void sample(
        self,
        double *data,
        Py_ssize_t length,
        double *mean,
        double *sum_abs,
        uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t i
        mean[0] = 0
        sum_abs[0] = 0
        for i in range(length):
            data[i] = rand_normal(self.mean, self.scale, seed)
            mean[0] += data[i]
            sum_abs[0] += fabs(data[i])

        mean[0] = mean[0] / length


cdef inline Py_ssize_t _max_exponent(
    Py_ssize_t n_timestep, Py_ssize_t kernel_size
) noexcept nogil:
    cdef Py_ssize_t max_exponent = <Py_ssize_t> floor(
        log2((n_timestep - 1) / <double> (kernel_size - 1))
    )
    if max_exponent < 0:
        max_exponent = 0
    max_exponent += 1

    return max_exponent


cdef class HydraAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_kernels
    cdef Py_ssize_t kernel_size
    cdef Py_ssize_t n_groups
    cdef KernelSampler kernel_sampler

    # Temporary buffers to store values while computing the convolution.
    cdef double *conv_values

    # Temporary buffers to store the min/max kernel values
    cdef double *min_values
    cdef double *max_values

    def __cinit__(
        self,
        Py_ssize_t n_groups,
        Py_ssize_t n_kernels,
        Py_ssize_t kernel_size,
        KernelSampler kernel_sampler,
    ):
        self.n_groups = n_groups
        self.n_kernels = n_kernels
        self.kernel_size = kernel_size
        self.kernel_sampler = kernel_sampler
        self.conv_values = NULL
        self.min_values = NULL
        self.max_values = NULL

    def __dealloc__(self):
        self._free()

    cdef void _free(self) noexcept nogil:
        if self.conv_values != NULL:
            free(self.conv_values)
            self.conv_values = NULL
        if self.min_values != NULL:
            free(self.min_values)
            self.min_values = NULL
        if self.max_values != NULL:
            free(self.max_values)
            self.max_values = NULL

    def __reduce__(self):
        return self.__class__, (
            self.n_groups, self.n_kernels, self.kernel_size, self.kernel_sampler
        )

    cdef int reset(self, TSArray X) noexcept nogil:
        self._free()
        self.conv_values = <double*> malloc(sizeof(double) * X.shape[2] * self.n_kernels)
        self.max_values = <double*> malloc(sizeof(double) * self.n_kernels)
        self.min_values = <double*> malloc(sizeof(double) * self.n_kernels)

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_groups

    # Each timeseries is represented by:
    #   n_groups * n_kernels * max_exponent * 2 (soft_max, hard_min)
    # attributes.
    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        cdef Py_ssize_t max_exponent = _max_exponent(X.shape[2], self.kernel_size)
        return self.get_n_attributess(X) * max_exponent * self.n_kernels * 2 * 1

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        cdef Hydra *hydra = <Hydra*> malloc(sizeof(Hydra))
        cdef Py_ssize_t i, j
        cdef double mean
        cdef double sum_abs
        cdef double *kernels = <double*> malloc(sizeof(double) * self.kernel_size * self.n_kernels)
        cdef double *kernel

        # Randomly sample and rescale n_kernels Hydra (2023), D (fit, Line 4)
        for i in range(self.n_kernels):
            kernel = kernels + i * self.kernel_size
            self.kernel_sampler.sample(kernel, self.kernel_size, &mean, &sum_abs, seed)

            # Hydra (2023), D (fit, Line 6)
            for j in range(self.kernel_size):
                kernel[j] = (kernel[j] - mean) / sum_abs

        hydra.kernels = kernels

        transient.dim = 0
        transient.attribute = hydra
        return 0

    # Restore a Hydra group from a numpy array
    cdef Py_ssize_t persistent_from_object(self, object obj, Attribute *attribute):
        dim, array = obj
        cdef Hydra *hydra = <Hydra*> malloc(sizeof(Hydra))

        # We will assume that array.shape[0] == hydra.n_kernels * hydra.kernel_size
        hydra.kernels = <double*> malloc(sizeof(double) * array.shape[0])
        cdef Py_ssize_t i

        # TODO: This can potentially be done without GIL.
        for i in range(array.shape[0]):
            hydra.kernels[i] = array[i]

        attribute.attribute = hydra
        attribute.dim = dim
        return 0

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return self.free_persistent(attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef Hydra *hydra
        if attribute.attribute != NULL:
            hydra = <Hydra*> attribute.attribute
            if hydra.kernels != NULL:
                free(hydra.kernels)
            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    # NOTE: We move ownership of the attribute here to a persistent attribute,
    # which will be freed by `free_persistent`.
    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        persistent.dim = transient.dim
        persistent.attribute = transient.attribute
        transient.attribute = NULL
        return 0

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef Py_ssize_t i
        cdef Hydra *hydra = <Hydra*> attribute.attribute
        cdef n = self.n_kernels * self.kernel_size
        data = np.empty(n, dtype=float)
        for i in range(n):
            data[i] = hydra.kernels[i]

        return attribute.dim, data

    cdef double transient_value(
        self, Attribute *attribute, TSArray X, Py_ssize_t sample
    ) noexcept nogil:
        return 0  # TODO

    cdef double persistent_value(
        self, Attribute *attribute, TSArray X, Py_ssize_t sample
    ) noexcept nogil:
        return self.transient_value(attribute, X, sample)

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        cdef Py_ssize_t max_exponent = _max_exponent(X.shape[2], self.kernel_size)
        cdef Py_ssize_t i, exponent, padding, dilation
        cdef Py_ssize_t kernel_attribute_offset

        cdef Py_ssize_t min_index, max_index
        cdef double min_value, max_value

        cdef Hydra *hydra = <Hydra*> attribute.attribute

        # Place the pointer inside correct attribute group, as given by attribute_id
        cdef Py_ssize_t attribute_offset = (
            attribute_id * self.n_kernels * max_exponent * 2
        )

        # Hydra (2023), D (transform, Line 5)
        for exponent in range(max_exponent):
            dilation = <Py_ssize_t> pow(2, exponent)
            padding = <Py_ssize_t> ((self.kernel_size - 1) * dilation) // 2

            # We store the output of the convolution in an array
            # of shape X.shape[2] * self.n_kernels
            for i in range(self.n_kernels):
                convolution_1d(
                    1,
                    dilation,
                    padding,
                    0.0,
                    hydra.kernels + i * self.kernel_size,
                    self.kernel_size,
                    &X[sample, attribute.dim, 0],
                    X.shape[2],
                    self.conv_values + i * X.shape[2],
                )

            memset(self.min_values, 0, sizeof(double) * self.n_kernels)
            memset(self.max_values, 0, sizeof(double) * self.n_kernels)

            # Find the timestep-wise kernel with the minimum and maximum
            # values. We iterate self.conv_values with a stride equal
            # to the number of timesteps.
            for i in range(X.shape[2]):
                find_min_max(
                    i,
                    X.shape[2],
                    self.conv_values,
                    self.n_kernels,
                    &min_index,
                    &min_value,
                    &max_index,
                    &max_value
                )

                # NOTE: min_index and max_index are bounded by self.n_kernels
                self.min_values[min_index] += 1
                self.max_values[max_index] += max_value

            # We allocate each attribute to an array with the following layout,
            # self.n_kernels * self.max_exponent * 2:
            #
            #        d=0            d=1
            # --------------- ---------------
            #   k=0     k=1     k=1     k=1
            # ------- ------- ------- -------
            # f=0 f=1 f=0 f=1 f=0 f=1 f=0 f=1
            # --- --- --- --- --- --- --- ---
            #
            # With self.n_groups such groups (one for each attribute_id)
            #
            # Here we move the pointer to first kernel of the d:th dilation
            # making sure that we account for the fact that each kernel
            # is descrived by two attributes.
            kernel_attribute_offset = attribute_offset + exponent * self.n_kernels * 2
            for i in range(self.n_kernels):
                # NOTE: *2 is the number of attributes (min/max)
                out[out_sample, kernel_attribute_offset + i * 2] = self.min_values[i]
                out[out_sample, kernel_attribute_offset + i * 2 + 1] = self.max_values[i]

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return self.transient_fill(
            attribute, X, sample, out, out_sample, out_attribute
        )

cdef void find_min_max(
    Py_ssize_t offset,
    Py_ssize_t stride,
    double* values,
    Py_ssize_t length,
    Py_ssize_t *min_index,
    double *min_value,
    Py_ssize_t *max_index,
    double *max_value
) noexcept nogil:
    cdef Py_ssize_t i
    cdef double value
    min_index[0] = -1
    max_index[0] = -1
    max_value[0] = -INFINITY
    min_value[0] = INFINITY

    for i in range(length):
        value = values[offset + i * stride]
        if value > max_value[0]:
            max_value[0] = value
            max_index[0] = i

        if value < min_value[0]:
            min_value[0] = value
            min_index[0] = i



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from wildboar.base import is_explainer
from wildboar.datasets import load_two_lead_ecg
from wildboar.explain import AmplitudeImportance, IntervalImportance, ShapeletImportance
from wildboar.tree import ShapeletTreeClassifier
from wildboar.utils._testing import (
    assert_exhaustive_parameter_checks,
    assert_parameter_checks,
)


@pytest.mark.parametrize(
    "explainer",
    [
        IntervalImportance(),
        AmplitudeImportance(),
        ShapeletImportance(),
    ],
)
def test_parameter_constrains(explainer):
    assert_exhaustive_parameter_checks(explainer)
    assert_parameter_checks(explainer)


@pytest.mark.parametrize(
    "explainer",
    [
        IntervalImportance(),
        AmplitudeImportance(),
        ShapeletImportance(),
    ],
)
def test_is_explainer(explainer):
    assert is_explainer(explainer)


def test_interval_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = IntervalImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    # fmt: off
    expected_scores = np.array([0.00438982, 0.01299385, 0.02809482, 0.2500439 , 0.21738367,
                                0.21071115, 0.00298507, 0.        , 0.        , 0.        ])
    # fmt: on
    assert_almost_equal(actual_scores, expected_scores)


def test_amplitude_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = AmplitudeImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    expected_scores = np.array([0.30640913, 0.20895522, 0.08165057, 0.0])

    assert_almost_equal(actual_scores, expected_scores)


def test_shapelet_importance():
    x_train, x_test, y_train, y_test = load_two_lead_ecg(False)
    clf = ShapeletTreeClassifier(n_shapelets=1000, random_state=123)
    clf.fit(x_train, y_train)

    imp = ShapeletImportance(scoring="accuracy", random_state=123)
    imp.fit(clf, x_test, y_test)

    actual_scores = imp.importances_.mean
    # fmt: off
    expected_scores = np.array(
        [0.38366989, 0.3046532 , 0.17559263, 0.37928007, 0.26777875,
         0.34855136, 0.24231782, 0.35381914, 0.13081651, 0.0403863 ]
    )
    # fmt: on
    assert_almost_equal(actual_scores, expected_scores)



# Authors: Isak Samsten
# License: BSD 3 clause

"""Utility functions."""

import os
import platform
import warnings

import numpy as np

from .validation import check_array, check_X_y

__all__ = [
    "check_array",
    "check_X_y",
]


class DependencyMissing:
    """
    Class to flag dependencies as missing.

    Parameters
    ----------
    e : ModuleNotFoundError, optional
        An exception.
    package : str, optional
        The package (installable using `pip`) required for the
        desired functionality.
    context : str, optional
        The context in which the error was raised.

    Examples
    --------
    In the following example we assume that `matplotlib` has not been installed
    by the user.

    >>> import numpy as np
    >>> from wildboar.utils import DependencyMissing
    >>> try:
    ...     import matplotlib.pylab as plt
    ... except ModuleNotFoundError as e:
    ...     plt = DependencyMissing(e, package="matplotlib", context="plot")
    >>> plt.plot(np.arange(10))
    Traceback (most recent call last):
    File "<stdin>", line 2, in <module>
    ModuleNotFoundError: No module named 'matplotlib'

    The above exception was the direct cause of the following exception:

    Traceback (most recent call last):
    ....
    ModuleNotFoundError: 'matplotlib' is required for 'plot', but not included
    in the default wildboar installation. Please run: `pip install matplotlib`
    to install the required package.
    """

    def __init__(self, e=None, *, package=None, context=None):
        if context is None:
            import inspect

            frm = inspect.stack()[1]
            if frm.function == "<module>":
                frm = inspect.getmodule(frm.frame)
                context = frm.__name__ if frm is not None else ""
            else:
                context = frm.function
        self.__package = package or e.name
        self.__context = context
        self.__e = e

    def __mro_entries__(self, bases):
        class Mock:
            def __init__(ignore_me, *args, **kwargs):
                _soft_dependency_error(
                    e=self.__e, package=self.__package, context=self.__context
                )

        return (Mock,)

    def __getattr__(self, name):
        _soft_dependency_error(
            e=self.__e, package=self.__package, context=self.__context
        )

    def __call__(self, *args, **kwds):
        _soft_dependency_error(
            e=self.__e, package=self.__package, context=self.__context
        )


def os_cache_path(dir):
    """
    Get the path to a operating system specific cache directory.

    Parameters
    ----------
    dir : str
        The sub-directory in the cache location.

    Returns
    -------
    str
        The cache path.
    """
    if platform.system() == "Windows":
        cache_dir = os.path.expandvars(r"%LOCALAPPDATA%\cache")
        return os.path.join(cache_dir, dir)
    elif platform.system() == "Linux":
        cache_dir = os.environ.get("XDG_CACHE_HOME")
        if not cache_dir:
            cache_dir = ".cache"
        return os.path.join(os.path.expanduser("~"), cache_dir, dir)
    elif platform.system() == "Darwin":
        return os.path.join(os.path.expanduser("~"), "Library", "Caches", dir)
    else:
        return os.path.join(os.path.expanduser("~"), ".cache", dir)


def _soft_dependency_error(e=None, package=None, context=None, warning=False):
    if e is None and package is None:
        raise ValueError("both e and package cant be None")

    if context is None:
        import inspect

        frm = inspect.stack()[1]
        if frm.function == "<module>":
            context = inspect.getmodule(frm.frame).__name__
        else:
            context = frm.function
    package = package or e.name
    msg = (
        f"'{package}' is required for '{context}', but not included in the default "
        f"wildboar installation. Please run: `pip install {package}` to install the "
        f"required package."
    )
    if warning:
        warnings.warn(msg, UserWarning)
    elif e is not None:
        raise ModuleNotFoundError(msg) from e
    else:
        raise ModuleNotFoundError(msg)


def _safe_jagged_array(lst):
    arr = np.empty(len(lst), dtype=object)
    arr[:] = lst
    return arr



# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

from libc.math cimport INFINITY, NAN, sqrt
from libc.stdlib cimport free, malloc

from ..utils cimport TSArray
from ..utils._fft cimport _pocketfft
from ..utils._stats cimport (
    IncStats,
    cumulative_mean_std,
    find_min,
    inc_stats_add,
    inc_stats_init,
    inc_stats_remove,
    inc_stats_variance,
)
from ._cdistance cimport (
    Metric,
    ScaledSubsequenceMetric,
    Subsequence,
    SubsequenceMetric,
    SubsequenceView,
)


cdef double EPSILON = 1e-10

cdef class ScaledMassSubsequenceMetric(ScaledSubsequenceMetric):
    cdef double *mean_x
    cdef double *std_x
    cdef double *dist_buffer
    cdef complex *x_buffer
    cdef complex *y_buffer

    def __init__(self):
        pass

    def __cinit__(self):
        self.mean_x = NULL
        self.std_x = NULL
        self.x_buffer = NULL
        self.y_buffer = NULL

    def __dealloc__(self):
        self.__free()

    def __reduce__(self):
        return self.__class__, ()

    cdef void __free(self) noexcept nogil:
        if self.mean_x != NULL:
            free(self.mean_x)
            self.mean_x = NULL
        if self.std_x != NULL:
            free(self.std_x)
            self.std_x = NULL
        if self.dist_buffer != NULL:
            free(self.dist_buffer)
            self.dist_buffer = NULL
        if self.x_buffer != NULL:
            free(self.x_buffer)
            self.x_buffer = NULL
        if self.y_buffer != NULL:
            free(self.y_buffer)
            self.y_buffer = NULL

    cdef int reset(self, TSArray X) noexcept nogil:
        self.__free()
        cdef Py_ssize_t n_timestep = X.shape[2]
        self.x_buffer = <complex*> malloc(sizeof(complex) * n_timestep)
        self.y_buffer = <complex*> malloc(sizeof(complex) * n_timestep)
        self.mean_x = <double*> malloc(sizeof(double) * n_timestep)
        self.std_x = <double*> malloc(sizeof(double) * n_timestep)
        self.dist_buffer = <double*> malloc(sizeof(double) * n_timestep)
        return 0

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=NULL,
    ) noexcept nogil:
        cumulative_mean_std(
            x,
            x_len,
            s_len,
            self.mean_x,
            self.std_x,
        )
        _mass_distance(
            x,
            x_len,
            s,
            s_len,
            s_mean,
            s_std,
            self.mean_x,
            self.std_x,
            self.x_buffer,
            self.y_buffer,
            self.dist_buffer,
        )
        return find_min(
            self.dist_buffer, x_len - s_len + 1, return_index
        )

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil:
        cumulative_mean_std(
            x,
            x_len,
            s_len,
            self.mean_x,
            self.std_x,
        )
        _mass_distance(
            x,
            x_len,
            s,
            s_len,
            s_mean,
            s_std,
            self.mean_x,
            self.std_x,
            self.x_buffer,
            self.y_buffer,
            distances,
        )

        cdef Py_ssize_t i, j
        j = 0
        for i in range(x_len - s_len + 1):
            if distances[i] <= threshold:
                distances[j] = distances[i]
                indicies[j] = i
                j += 1
        return j

    cdef void _distance_profile(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double *dp,
    ) noexcept nogil:
        cumulative_mean_std(
            x,
            x_len,
            s_len,
            self.mean_x,
            self.std_x,
        )
        _mass_distance(
            x,
            x_len,
            s,
            s_len,
            s_mean,
            s_std,
            self.mean_x,
            self.std_x,
            self.x_buffer,
            self.y_buffer,
            dp,
        )

cdef void _mass_distance(
    const double *x,
    Py_ssize_t x_length,
    const double *y,
    Py_ssize_t y_length,
    double mean,
    double std,
    double *mean_x,     # length x_length - y_length + 1
    double *std_x,      # length x_length - y_length + 1
    complex *y_buffer,  # length x_length
    complex *x_buffer,  # length x_length
    double *dist,       # length x_length - y_length + 1
) noexcept nogil:
    cdef Py_ssize_t i
    cdef double z
    for i in range(x_length):
        if i < y_length:
            y_buffer[i] = y[y_length - i - 1]
        else:
            y_buffer[i] = 0
        x_buffer[i] = x[i]

    _pocketfft.fft(y_buffer, x_length, 1.0)
    _pocketfft.fft(x_buffer, x_length, 1.0)
    for i in range(x_length):
        x_buffer[i] *= y_buffer[i]
    _pocketfft.ifft(x_buffer, x_length, 1.0 / x_length)

    for i in range(x_length - y_length + 1):
        if (
            std_x[i] <= EPSILON
            and not std <= EPSILON
            or std <= EPSILON
            and not std_x[i] <= EPSILON
        ):
            dist[i] = sqrt(y_length)
        elif std_x[i] <= EPSILON and std <= EPSILON:
            dist[i] = 0
        else:
            z = x_buffer[i + y_length - 1].real
            z = 2 * (y_length - (z - y_length * mean_x[i] * mean) / (std_x[i] * std))
            if z < EPSILON:
                dist[i] = 0
            else:
                dist[i] = sqrt(z)



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, NAN, log2
from libc.stdlib cimport calloc, free, malloc
from libc.string cimport memcpy, memset
from numpy cimport uint32_t

from ..distance._cdistance cimport MetricList
from ..utils cimport TSArray
from ..utils._misc cimport argsort, safe_realloc
from ..utils._rand cimport (
    RAND_R_MAX,
    VoseRand,
    rand_int,
    vose_rand_free,
    vose_rand_init,
    vose_rand_int,
    vose_rand_precompute,
)


cdef struct Pivot:
    double** data
    Py_ssize_t metric
    Py_ssize_t length
    Py_ssize_t n_branches


cdef struct Split:
    Py_ssize_t* split_point  # n_split
    Py_ssize_t *pivot  # n_split + 1
    Py_ssize_t metric
    Py_ssize_t n_split
    double impurity_improvement
    double *child_impurity  # n_split + 1


cdef void free_split(Split *split) noexcept nogil:
    if split.split_point != NULL:
        free(split.split_point)
        split.split_point = NULL

    if split.pivot != NULL:
        free(split.pivot)
        split.pivot = NULL

    if split.child_impurity != NULL:
        free(split.child_impurity)
        split.child_impurity = NULL

cpdef Tree _make_tree(
    list metrics,
    Py_ssize_t n_labels,
    Py_ssize_t max_depth,
    object branch,
    list pivots,
    object value,
):
    cdef Tree tree = Tree(metrics, n_labels, capacity=len(pivots) + 1)
    tree._max_depth = max_depth
    tree._node_count = len(pivots)
    cdef Py_ssize_t i, j, k
    cdef Pivot *pivot
    value = value.reshape(-1)
    for i in range(tree._node_count):
        if pivots[i] is not None:
            pivot = <Pivot*> malloc(sizeof(Pivot))
            pivot.metric = pivots[i][0]
            arr = pivots[i][1]
            pivot.n_branches = arr.shape[0]
            pivot.length = arr.shape[1]
            pivot.data = <double**> malloc(sizeof(double*) * pivot.n_branches)
            for j in range(pivot.n_branches):
                pivot.data[j] = <double*> malloc(sizeof(double) * pivot.length)
                for k in range(pivot.length):
                    pivot.data[j][k] = arr[j, k]

            tree._pivots[i] = pivot
        else:
            tree._pivots[i] = NULL

        for j in range(n_labels):
            tree._branches[j][i] = branch[j, i]

    for i in range(tree._node_count * n_labels):
        tree._values[i] = value[i]

    return tree


cdef class Tree:

    cdef Py_ssize_t _max_depth
    cdef Py_ssize_t _node_count
    cdef Py_ssize_t **_branches
    cdef Pivot **_pivots
    cdef double *_values
    cdef Py_ssize_t _capacity
    cdef Py_ssize_t _n_labels
    cdef MetricList metrics

    def __cinit__(
        self,
        list metrics,
        Py_ssize_t n_labels,
        Py_ssize_t capacity=10
    ):
        self._node_count = 0
        self._capacity = capacity
        self._n_labels = n_labels

        self._branches = <Py_ssize_t**>malloc(sizeof(Py_ssize_t*) * self._n_labels)
        cdef Py_ssize_t i
        for i in range(self._n_labels):
            self._branches[i] = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * capacity)

        self._pivots = <Pivot**> malloc(sizeof(Pivot*) * capacity)
        self._values = <double*> malloc(sizeof(double) * capacity * n_labels)
        self.metrics = MetricList(metrics)

    def __reduce__(self):
        return _make_tree, (
            self.metrics.py_list,
            self._n_labels,
            self._max_depth,
            self.branch,
            self.pivot,
            self.value,
        )

    cdef Py_ssize_t add_branch_node(
        self,
        Py_ssize_t parent,
        Py_ssize_t branch,
        Py_ssize_t n_samples,
        double n_weighted_samples,
        TSArray X,
        Py_ssize_t *pivots,
        Py_ssize_t metric,
        Py_ssize_t n_split,
    ) noexcept nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() < 0:
                return -1

        cdef Pivot *pivot = <Pivot*> malloc(sizeof(Pivot))
        pivot.n_branches = n_split + 1
        pivot.length = X.shape[2]
        pivot.metric = metric
        pivot.data = <double**> malloc(sizeof(double*) * pivot.n_branches)
        cdef Py_ssize_t i
        for i in range(pivot.n_branches):
            pivot.data[i] = <double*> malloc(sizeof(double) * X.shape[2])
            memcpy(
                pivot.data[i], &X[pivots[i], 0, 0], sizeof(double) * X.shape[2]
            )

        self._pivots[node_id] = pivot
        if parent != -1:
            self._branches[branch][parent] = node_id

        self._node_count += 1
        return node_id

    cdef Py_ssize_t add_leaf_node(self, Py_ssize_t parent, Py_ssize_t branch) noexcept nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() < 0:
                return -1

        if parent != -1:
            self._branches[branch][parent] = node_id
        cdef Py_ssize_t i
        for i in range(self._n_labels):
            self._branches[i][node_id] = -1

        self._pivots[node_id] = NULL
        self._node_count += 1
        return node_id

    cdef void set_leaf_value(
        self,
        Py_ssize_t node_id,
        Py_ssize_t label,
        double value
    ) noexcept nogil:
        self._values[label + node_id * self._n_labels] = value

    cdef Py_ssize_t _increase_capacity(self) noexcept nogil:
        cdef Py_ssize_t new_capacity = self._node_count * 2
        cdef Py_ssize_t i
        for i in range(self._n_labels):
            safe_realloc(<void**> &self._branches[i], sizeof(Py_ssize_t) * new_capacity)
        safe_realloc(<void**> &self._pivots, sizeof(Pivot*) * new_capacity)
        safe_realloc(<void**> &self._values, sizeof(double) * new_capacity * self._n_labels)
        self._capacity = new_capacity
        return 0

    @property
    def branch(self):
        branches = np.zeros((self._n_labels, self._node_count), dtype=np.intp)
        for i in range(self._n_labels):
            for j in range(self._node_count):
                branches[i, j] = self._branches[i][j]
        return branches

    @property
    def pivot(self):
        pivots = []
        cdef Pivot *pivot
        for i in range(self._node_count):
            pivot = self._pivots[i]
            if pivot != NULL:
                arr = np.empty((pivot.n_branches, pivot.length), dtype=np.double)
                for j in range(pivot.n_branches):
                    for k in range(pivot.length):
                        arr[j, k] = self._pivots[i].data[j][k]
                pivots.append((pivot.metric, arr))
            else:
                pivots.append(None)

        return pivots

    @property
    def value(self):
        cdef object arr = np.empty(self._node_count * self._n_labels, dtype=float)
        cdef Py_ssize_t i
        for i in range(self._n_labels * self._node_count):
            arr[i] = self._values[i]
        return arr.reshape(self._node_count, self._n_labels)

    @property
    def max_depth(self):
        return self._max_depth

    def apply(self, TSArray X):
        cdef Py_ssize_t[:] out = np.zeros((X.shape[0],), dtype=np.intp)
        cdef Py_ssize_t node_index
        cdef Py_ssize_t i
        cdef Py_ssize_t branch
        cdef Pivot *pivot

        with nogil:
            for i in range(self.metrics.size):
                self.metrics.reset(i, X, X)

            for i in range(X.shape[0]):
                node_index = 0
                while self._branches[0][node_index] != -1:
                    pivot = self._pivots[node_index]
                    branch = find_min_branch(
                        self.metrics, pivot, &X[i, 0, 0]
                    )
                    node_index = self._branches[branch][node_index]
                out[i] = node_index

        return out.base

    def predict(self, TSArray X):
        return np.take(self.value, self.apply(X), axis=0, mode="clip")


cdef Py_ssize_t find_min_branch(
    MetricList metrics,
    Pivot *pivot,
    const double *sample,
) noexcept nogil:
    cdef double dist
    cdef double min_dist = INFINITY
    cdef Py_ssize_t min_branch = -1

    cdef Py_ssize_t i
    for i in range(pivot.n_branches):
        dist = metrics._distance(
            pivot.metric, sample, pivot.length, pivot.data[i], pivot.length
        )
        if dist < min_dist:
            min_dist = dist
            min_branch = i
    return min_branch


cdef class Criterion:

    cdef const Py_ssize_t[:] labels

    cdef Py_ssize_t n_labels
    cdef double weighted_n_total
    cdef double *weighted_n_branch
    cdef double *weighted_label_count
    cdef double *weighted_label_branch_count
    cdef Py_ssize_t *label_count

    cdef Py_ssize_t start
    cdef Py_ssize_t end
    cdef Py_ssize_t *samples
    cdef const double[:] sample_weight

    def __cinit__(self, const Py_ssize_t[:] y, Py_ssize_t n_labels):
        self.labels = y
        self.n_labels = n_labels
        self.weighted_label_count = <double*> calloc(n_labels, sizeof(double))
        self.weighted_n_branch = <double*> calloc(n_labels, sizeof(double))
        self.weighted_label_branch_count = <double*> calloc(
            n_labels * n_labels, sizeof(double)
        )
        self.label_count = <Py_ssize_t*> calloc(n_labels, sizeof(Py_ssize_t))

    cdef void init(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *samples,
        const double[:] sample_weight,
    ) noexcept nogil:
        self.start = start
        self.end = end
        self.samples = samples
        self.sample_weight = sample_weight

        memset(self.weighted_label_count, 0, sizeof(double) * self.n_labels)
        memset(self.label_count, 0, sizeof(Py_ssize_t) * self.n_labels)
        self.weighted_n_total = 0
        cdef Py_ssize_t i, j
        cdef double w = 1.0
        for i in range(start, end):
            j = self.samples[i]
            if self.sample_weight is not None:
                w = self.sample_weight[j]

            self.weighted_n_total += w
            self.weighted_label_count[self.labels[j]] += w
            self.label_count[self.labels[j]] += 1

    cdef void reset(self, double *samples_branch) noexcept nogil:
        cdef Py_ssize_t i, j
        cdef Py_ssize_t label, branch
        cdef double w = 1.0
        memset(
            self.weighted_label_branch_count,
            0,
            sizeof(double) * self.n_labels * self.n_labels
        )
        memset(self.weighted_n_branch, 0, sizeof(double) * self.n_labels)
        for i in range(self.start, self.end):
            j = self.samples[i]
            label = self.labels[j]
            branch = (<Py_ssize_t>samples_branch[i])
            if self.sample_weight is not None:
                w = self.sample_weight[j]

            self.weighted_n_branch[branch] += w
            self.weighted_label_branch_count[label + branch * self.n_labels] += w

    cdef double impurity(self) noexcept nogil:
        return NAN

    cdef void child_impurity(self, double *branches, Py_ssize_t n_branches) noexcept nogil:
        pass

    cdef double proxy_impurity(self, double* branches, Py_ssize_t n_branches) noexcept nogil:
        self.child_impurity(branches, n_branches)
        cdef double impurity = 0.0
        for i in range(n_branches):
            impurity -= self.weighted_n_branch[i] * branches[i]
        return impurity

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double *child_impurity,
        Py_ssize_t n_branches,
        double n_weighted_samples,
    ) noexcept nogil:
        cdef Py_ssize_t i
        for i in range(n_branches):
            impurity_parent -= self.weighted_n_branch[i] / self.weighted_n_total * child_impurity[i]
        return self.weighted_n_total / n_weighted_samples * impurity_parent


cdef class GiniCriterion(Criterion):

    cdef double impurity(self) noexcept nogil:
        cdef double sq_count = 0.0
        cdef double c
        cdef Py_ssize_t i
        for i in range(self.n_labels):
            c = self.weighted_label_count[i]
            sq_count += c * c
        return 1.0 - sq_count / (self.weighted_n_total * self.weighted_n_total)

    cdef void child_impurity(self, double *branches, Py_ssize_t n_branches) noexcept nogil:
        cdef Py_ssize_t label, branch
        cdef double v
        for branch in range(n_branches):
            branches[branch] = 0.0
            for label in range(self.n_labels):
                v = self.weighted_label_branch_count[label + branch * self.n_labels]
                branches[branch] += v * v

        for branch in range(n_branches):
            if self.weighted_n_branch[branch] > 0:
                branches[branch] = 1 - branches[branch] / (
                    self.weighted_n_branch[branch] * self.weighted_n_branch[branch]
                )


cdef class EntropyCriterion(Criterion):

    cdef double impurity(self) noexcept nogil:
        cdef double c
        cdef double entropy = 0
        cdef Py_ssize_t i
        for i in range(self.n_labels):
            c = self.weighted_label_count[i]
            if c > 0:
                c /= self.weighted_n_total
                entropy -= c * log2(c)
        return entropy

    cdef void child_impurity(self, double *branches, Py_ssize_t n_branches) noexcept nogil:
        cdef double v
        cdef Py_ssize_t label, branch
        for branch in range(n_branches):
            branches[branch] = 0
            for label in range(self.n_labels):
                v = self.weighted_label_branch_count[label + branch * self.n_labels]
                if v > 0:
                    v /= self.weighted_n_branch[branch]
                    branches[branch] -= v * log2(v)


cdef class MetricSampler:
    cdef VoseRand vr
    cdef Py_ssize_t n_measures

    def __cinit__(self, Py_ssize_t n_measures, const double[::1] weights=None):
        self.n_measures = n_measures
        if weights is not None:
            vose_rand_init(&self.vr, weights.shape[0])
            vose_rand_precompute(&self.vr, &weights[0])

    def __dealloc__(self):
        vose_rand_free(&self.vr)

    cdef Py_ssize_t sample(
        self,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        uint32_t *seed
    ) noexcept nogil:
        pass

cdef class UniformMetricSampler(MetricSampler):

    cdef Py_ssize_t sample(
        self,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        uint32_t *seed
    ) noexcept nogil:
        return rand_int(0, self.n_measures, seed)


cdef class WeightedMetricSampler(MetricSampler):

    cdef Py_ssize_t sample(
        self,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        uint32_t *seed
    ) noexcept nogil:
        return vose_rand_int(&self.vr, seed)


cdef class PivotSampler:

    cdef Py_ssize_t sample(
        self,
        const Py_ssize_t[:] labels,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Py_ssize_t label,
        Py_ssize_t *label_count,
        uint32_t *seed,
    ) noexcept nogil:
        pass


cdef class LabelPivotSampler(PivotSampler):

    cdef Py_ssize_t sample(
        self,
        const Py_ssize_t[:] labels,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Py_ssize_t label,
        Py_ssize_t *label_count,
        uint32_t *seed,
    ) noexcept nogil:
        cdef Py_ssize_t n_labels = rand_int(0, label_count[label], seed)
        cdef Py_ssize_t i, j
        cdef Py_ssize_t label_index = 0
        for i in range(n_samples):
            j = samples[i]
            if label_index == n_labels:
                return j

            if labels[j] == label:
                label_index += 1

        # This code is unreachable
        return -1

cdef class UniformPivotSampler(PivotSampler):

    cdef Py_ssize_t sample(
        self,
        const Py_ssize_t[:] labels,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Py_ssize_t label,
        Py_ssize_t *label_count,
        uint32_t *seed,
    ) noexcept nogil:
        return samples[rand_int(0, n_samples, seed)]

cdef class TreeBuilder:

    cdef Py_ssize_t n_attributes
    cdef Py_ssize_t max_depth
    cdef Py_ssize_t min_samples_split
    cdef Py_ssize_t min_samples_leaf
    cdef double min_impurity_decrease

    cdef Py_ssize_t n_samples  # no samples with non-zero weight
    cdef double n_weighted_samples

    cdef Py_ssize_t *samples
    cdef const double[:] sample_weights

    cdef double *samples_branch_buffer
    cdef double *samples_branch
    cdef Py_ssize_t *pivot_buffer
    cdef Py_ssize_t *branch_count

    cdef TSArray X
    cdef MetricList metrics
    cdef readonly Tree tree
    cdef Criterion criterion
    cdef MetricSampler metric_sampler
    cdef PivotSampler pivot_sampler
    cdef uint32_t seed

    def __cinit__(
        self,
        TSArray X,
        const double[:] sample_weights,
        PivotSampler pivot_sampler,
        MetricSampler metric_sampler,
        Criterion criterion,
        Tree tree,
        object random_state,
        Py_ssize_t n_attributes=1,
        Py_ssize_t max_depth=2**10,
        Py_ssize_t min_samples_split=2,
        Py_ssize_t min_samples_leaf=1,
        double min_impurity_decrease=0.0
    ):
        self.max_depth = max_depth
        self.n_attributes = n_attributes
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_impurity_decrease = min_impurity_decrease

        self.X = X
        self.criterion = criterion
        self.tree = tree
        self.metric_sampler = metric_sampler
        self.pivot_sampler = pivot_sampler
        self.metrics = tree.metrics
        self.seed = random_state.randint(0, RAND_R_MAX)

        self.sample_weights = sample_weights
        self.samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.X.shape[0])
        self.samples_branch = <double*> malloc(sizeof(double) * self.X.shape[0])
        self.samples_branch_buffer = <double*> malloc(sizeof(double) * self.X.shape[0])

        self.pivot_buffer = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.criterion.n_labels)
        self.branch_count = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.criterion.n_labels)

        if (
            self.samples == NULL or
            self.samples_branch == NULL or
            self.samples_branch_buffer == NULL or
            self.pivot_buffer == NULL or
            self.branch_count == NULL
        ):
            raise MemoryError()

        cdef Py_ssize_t i, j
        j = 0
        self.n_weighted_samples = 0.0
        for i in range(self.X.shape[0]):
            if sample_weights is None or sample_weights[i] != 0.0:
                self.samples[j] = i
                j += 1
                if sample_weights is not None:
                    self.n_weighted_samples += sample_weights[i]
                else:
                    self.n_weighted_samples += 1.0

        self.n_samples = j

        for i in range(self.metrics.size):
            self.metrics.reset(i, self.X, self.X)

    def __dealloc__(self):
        free(self.samples)
        free(self.samples_branch_buffer)
        free(self.pivot_buffer)
        free(self.samples_branch)
        free(self.branch_count)

    def build_tree(self):
        cdef Py_ssize_t max_depth = 0
        with nogil:
            self._build_tree(
                0,
                self.n_samples,
                0,
                -1,
                0,
                NAN,
                &max_depth,
            )
        self.tree._max_depth = max_depth

    cdef void _build_tree(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t depth,
        Py_ssize_t parent,
        Py_ssize_t branch,
        double impurity,
        Py_ssize_t *max_depth
    ) noexcept nogil:
        if depth > max_depth[0]:
            max_depth[0] = depth

        self.criterion.init(start, end, self.samples, self.sample_weights)
        cdef Py_ssize_t n_node_samples = end - start
        cdef bint is_leaf = (
            depth >= self.max_depth
            or n_node_samples < self.min_samples_split
            or n_node_samples < 2 * self.min_samples_leaf
        )

        cdef Py_ssize_t i
        if is_leaf:
            self._new_leaf_node(parent, branch)
            return

        if parent < 0:
            impurity = self.criterion.impurity()

        cdef Py_ssize_t j
        cdef Py_ssize_t node_id
        cdef Py_ssize_t current_split
        cdef Py_ssize_t split_start
        cdef double current_weight = 1.0
        cdef double p
        cdef Split split = self._split(start, end, impurity)
        is_leaf = (
            split.n_split < 1 or
            split.impurity_improvement <= self.min_impurity_decrease
        )

        if not is_leaf:
            node_id = self.tree.add_branch_node(
                parent,
                branch,
                n_node_samples,
                self.criterion.weighted_n_total,
                self.X,
                split.pivot,
                split.metric,
                split.n_split,
            )
            split_start = start
            for current_split in range(split.n_split):
                self._build_tree(
                    split_start,
                    split.split_point[current_split],
                    depth + 1,
                    node_id,
                    current_split,
                    split.child_impurity[current_split],
                    max_depth,
                )
                split_start = split.split_point[current_split]

            self._build_tree(
                split.split_point[split.n_split - 1],
                end,
                depth + 1,
                node_id,
                split.n_split,
                split.child_impurity[split.n_split],
                max_depth,
            )
        else:
            self._new_leaf_node(parent, branch)

        free_split(&split)

    cdef void _new_leaf_node(self, Py_ssize_t parent, Py_ssize_t branch) noexcept nogil:
        cdef Py_ssize_t node_id = self.tree.add_leaf_node(parent, branch)
        cdef Py_ssize_t i
        for i in range(self.criterion.n_labels):
            self.tree.set_leaf_value(
                node_id,
                i,
                self.criterion.weighted_label_count[i] / self.criterion.weighted_n_total
            )

    cdef Split _split(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        double impurity_parent,
    ) noexcept nogil:
        cdef Py_ssize_t n_samples = end - start
        cdef Py_ssize_t n_branches = 0
        cdef Py_ssize_t i, j
        for i in range(self.criterion.n_labels):
            if self.criterion.label_count[i] > 0:
                n_branches += 1

        cdef Py_ssize_t label
        cdef Py_ssize_t pivot_index
        cdef Py_ssize_t best_metric = -1
        cdef double impurity
        cdef double best_impurity
        cdef Split split

        # Abort early if the samples of this branch has a single label
        if n_branches <= 1:
            split.n_split = 0
            split.impurity_improvement = INFINITY
            split.split_point = NULL
            split.pivot = NULL
            split.child_impurity = NULL
            split.metric = -1
            return split
        else:
            split.pivot = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.criterion.n_labels)
            split.split_point = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.criterion.n_labels)
            split.child_impurity = <double*> malloc(sizeof(double) * self.criterion.n_labels)
            best_impurity = -INFINITY
            # TODO: we should not stop until a split with improvement is found.
            for _ in range(self.n_attributes):
                pivot_index = 0
                for label in range(self.criterion.n_labels):
                    if self.criterion.label_count[label] > 0:
                        split.pivot[pivot_index] = self.pivot_sampler.sample(
                            self.criterion.labels,
                            self.samples + start,
                            n_samples,
                            label,
                            self.criterion.label_count,
                            &self.seed,
                        )
                        pivot_index += 1

                split.metric = self.metric_sampler.sample(
                    self.samples + start, n_samples, &self.seed,
                )
                self._partition_pivots(
                    start, end, split.pivot, split.metric, n_branches
                )
                self.criterion.reset(self.samples_branch)
                impurity = self.criterion.proxy_impurity(split.child_impurity, n_branches)
                if impurity > best_impurity:
                    best_impurity = impurity
                    best_metric = split.metric
                    memcpy(
                        self.pivot_buffer,
                        split.pivot,
                        sizeof(Py_ssize_t) * self.criterion.n_labels,
                    )
                    memcpy(
                        self.samples_branch_buffer + start,
                        self.samples_branch + start,
                        sizeof(double) * n_samples,
                    )

            # Restore the partitioning with the lowest impurity score
            memcpy(
                self.samples_branch + start,
                self.samples_branch_buffer + start,
                sizeof(double) * n_samples,
            )

            argsort(self.samples_branch + start, self.samples + start, n_samples)
            self.criterion.reset(self.samples_branch)
            self.criterion.child_impurity(split.child_impurity, n_branches)

            # Restore the pivots with the lowest impurity score
            memcpy(split.pivot, self.pivot_buffer, sizeof(Py_ssize_t) * self.criterion.n_labels)
            split.metric = best_metric

            # Shift unused pivots to the end. The number of used pivots might be lower
            # than the number of possible branches. We compute the number of splits
            # in the _find_split_points-method. The number of branches is the number
            # of splits + 1.
            memset(self.branch_count, 0, sizeof(Py_ssize_t) * self.criterion.n_labels)
            for i in range(start, end):
                self.branch_count[<Py_ssize_t> self.samples_branch[i]] += 1

            j = 0
            for i in range(n_branches):
                if self.branch_count[i] > 0:
                    split.pivot[j] = split.pivot[i]
                    split.child_impurity[j] = split.child_impurity[i]
                    j += 1
                else:
                    split.pivot[i] = -1
                    split.child_impurity[i] = -1

            self._find_split_points(start, end, &split)
            split.impurity_improvement = self.criterion.impurity_improvement(
                impurity_parent,
                split.child_impurity,
                split.n_split + 1,
                self.n_weighted_samples,
            )

            return split

    # Identify the indices where the values of the samples_branch array changes.
    cdef Py_ssize_t _find_split_points(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Split *split
    ) noexcept nogil:
        cdef Py_ssize_t i, current_split
        current_split = 0
        for i in range(start + 1, end):
            if self.samples_branch[i] != self.samples_branch[i - 1]:
                split.split_point[current_split] = i
                current_split += 1

        # Set the number of splits to the number of actual split points.
        split.n_split = current_split

    # Partition the samples to the branch of the pivot with the smallest distance
    cdef void _partition_pivots(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *pivots,
        Py_ssize_t metric,
        Py_ssize_t n_branches
    ) noexcept nogil:
        cdef Py_ssize_t i, j
        cdef Py_ssize_t pivot
        cdef Py_ssize_t min_pivot = -1
        cdef double min_dist = INFINITY
        cdef double dist
        for i in range(start, end):
            j = self.samples[i]
            min_dist = INFINITY
            for pivot in range(n_branches):
                dist = self.metrics.distance(
                    metric, self.X, pivots[pivot], self.X, j, 0
                )

                if dist < min_dist:
                    min_dist = dist
                    min_pivot = pivot
            self.samples_branch[i] = min_pivot



# Authors: Isak Samsten
# License: BSD 3 clause

"""Tree-based estimators for classification and regression."""

from ._ptree import ProximityTreeClassifier
from ._tree import (
    ExtraShapeletTreeClassifier,
    ExtraShapeletTreeRegressor,
    IntervalTreeClassifier,
    IntervalTreeRegressor,
    PivotTreeClassifier,
    RocketTreeClassifier,
    RocketTreeRegressor,
    ShapeletTreeClassifier,
    ShapeletTreeRegressor,
)

__all__ = [
    "ShapeletTreeClassifier",
    "ShapeletTreeRegressor",
    "ExtraShapeletTreeClassifier",
    "ExtraShapeletTreeRegressor",
    "RocketTreeClassifier",
    "RocketTreeRegressor",
    "IntervalTreeClassifier",
    "IntervalTreeRegressor",
    "PivotTreeClassifier",
    "ProximityTreeClassifier",
]



# Authors: Isak Samsten
# License: BSD 3 clause
import math
import warnings

import numpy as np

from ..distance import matrix_profile
from ..utils.decorators import array_or_scalar
from ..utils.validation import check_array


@array_or_scalar()
def segment(
    x=None,
    *,
    mpi=None,
    n_segments=1,
    window=0.1,
    exclude=0.2,
    boundry=1.0,
    return_arc_curve=False,
):
    """Find change regimes in a time series.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep) or (n_timestep, ), optional
        The time series. If x is given, the matrix profile of x is computed.

    mpi : array-like of shape (n_samples, profile_size) or (profile_size), optional
        The matrix profile index. Must be given unless x is given.

    n_segments : int, optional
        The number of segmentations to identify

    window : int or float, optional
        The window size. The window parameter is ignored if mpi is not None.

        - if float, a fraction of n_timestep
        - if int, the exact window size

    exclude : float, optional
        The self-join exclusion for the matrix profile. Ignored if `mpi` is given.

    boundry : float, optional
        Ignore the region around each identified segmentation. Expressed as a
        fraction of `window`.

    return_arc_curve : bool, optional
        Return the arc curve.

    Returns
    -------
    segments : ndarray of shape (n_samples, n_segments), (n_segments) or int
        The start index of a segment
    arc_curves : ndarray of shape (n_samples, profile_size) or (profile_size, )
        The arc curves

    See Also
    --------
    wildboar.distance.matrix_profile : compute the matrix profile

    References
    ----------
    Gharghabi, Shaghayegh, et al. (2017)
        Matrix profile VIII: domain agnostic online semantic segmentation at superhuman
        performance levels. In proceedings of International Conference on Data Mining
    """
    if x is None and mpi is None:
        raise ValueError("Either x or mpi must be given")
    elif x is not None:
        if mpi is not None:
            raise ValueError("Both x and mpi cannot be given")

        x = check_array(np.atleast_2d(x), allow_3d=False)
        _, mpi = matrix_profile(x, window=window, exclude=exclude, return_index=True)
        mpi = np.atleast_2d(mpi)
    else:
        mpi = check_array(np.atleast_2d(mpi))

    boundry = math.ceil(window * boundry)
    segments = -np.ones((mpi.shape[0], n_segments), dtype=np.intp)
    if return_arc_curve:
        arc_curves = np.empty(mpi.shape, dtype=np.double)

    index = np.arange(mpi.shape[-1])
    arc_curve_normalize = 2 * index * (mpi.shape[-1] - index) / mpi.shape[-1]
    arc_curve_normalize[0] = 10e-10
    for i in range(mpi.shape[0]):
        arc_curve = np.minimum(
            np.cumsum(
                np.bincount(np.minimum(index, mpi[i]), minlength=mpi.shape[-1])
                - np.bincount(np.maximum(index, mpi[i]), minlength=mpi.shape[-1])
            )
            / arc_curve_normalize,
            1,
        )
        arc_curve[:boundry] = 1.0
        arc_curve[-boundry:] = 1.0

        if return_arc_curve:
            arc_curves[i, :] = arc_curve

        for j in range(n_segments):
            segments[i, j] = np.argmin(arc_curve)
            if arc_curve[segments[i, j]] == 1.0:
                warnings.warn(
                    f"No more segments for sample={i} (segment={j}). All remaining"
                    f"segments are invalid and set to -1.",
                    UserWarning,
                )
                break
            start = max(segments[i, j] - boundry, 0)
            end = min(segments[i, j] + boundry, arc_curve.shape[0])
            arc_curve[start:end] = 1.0

    if return_arc_curve:
        return segments, arc_curves
    else:
        return segments



# Authors: Isak Samsten
# License: BSD 3 clause

"""Utilities for variable length series."""
import numpy as np


def _float_to_raw_int_bits(value):
    return np.array(value, dtype="f").view(dtype="I")


def _int_bits_to_float(bits):
    return np.array(bits, dtype="I").view(dtype="f")


eos = EoS = EOS = _int_bits_to_float(0x7F800009).item()
_END_OF_SERIES_MASK = np.array(0x0000000F, dtype=np.uint64)


def is_end_of_series(x):
    """
    Test element-wise for EoS and return result as a boolean array.

    Parameters
    ----------
    x : ndarray
        Input array.

    Returns
    -------
    ndarray
        Boolean indicator array.
    """
    return np.logical_and(
        np.isnan(x),
        np.bitwise_and(_float_to_raw_int_bits(x), _END_OF_SERIES_MASK) == 9,
    )


def is_variable_length(x):
    """
    Test if time-series is variable length.

    Parameters
    ----------
    x : time-series
        The input.

    Returns
    -------
    bool
        True if time series contains EoS.
    """
    return is_end_of_series(x).any()


def get_variable_length(x):
    """
    Return the length of each time-series.

    Parameters
    ----------
    x : time-series
        The input.

    Returns
    -------
    float or ndarray
        The lenght of the time series.

    Examples
    --------
    >>> from wildboar.utils.variable_len import get_variable_length
    >>> x = np.array(
    ...     [
    ...         [[1, 2, 3, eos], [1, 2, eos, eos], [1, 2, 3, 4]],
    ...         [[1, 2, 3, eos], [1, 2, eos, eos], [1, eos, 3, 4]],
    ...     ]
    ... )
    >>> get_variable_length(x)
    [[3 2 4]
     [3 2 4]]

    >>> get_variable_length([1, 2, eos, eos])
    2

    >>> get_variable_length([[1, 2, eos, eos]])
    [2]
    """
    eos = is_end_of_series(x)
    if eos.ndim == 1:
        argmax = np.r_[False, eos].argmax() - 1
        return eos.shape[0] if argmax < 0 else argmax
    else:
        if eos.ndim == 2:
            f = np.broadcast_to(False, shape=(eos.shape[0], 1))
        elif eos.ndim == 3:
            f = np.broadcast_to(False, shape=(eos.shape[0], eos.shape[1], 1))
        else:
            raise ValueError(
                f"Found array with {eos.ndim} dimensions but expected <= 3"
            )

        argmax = np.c_[f, eos].argmax(axis=-1) - 1
        argmax[argmax == -1] = eos.shape[-1]
        return argmax



# Authors: Isak Samsten
# License: BSD 3 clause
import numbers

import numpy as np
from joblib import effective_n_jobs
from sklearn.pipeline import make_pipeline, make_union
from sklearn.utils import check_random_state

from ..datasets.preprocess import SparseScaler
from ..transform import (
    CastorTransform,
    DiffTransform,
    DilatedShapeletTransform,
    RandomShapeletTransform,
)
from ._transform import TransformRidgeClassifierCV, TransformRidgeCV


class RandomShapeletClassifier(TransformRidgeClassifierCV):
    """A classifier that uses random shapelets."""

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0.1,
        max_shapelet_size=1.0,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        class_weight=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def _get_transform(self, random_state):
        return RandomShapeletTransform(
            self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


class RandomShapeletRegressor(TransformRidgeCV):
    """A regressor that uses random shapelets."""

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        min_shapelet_size=0.1,
        max_shapelet_size=1.0,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        normalize=False,
        scoring=None,
        cv=None,
        gcv_mode=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            gcv_mode=gcv_mode,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def _get_transform(self, random_state):
        return RandomShapeletTransform(
            self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


class DilatedShapeletClassifier(TransformRidgeClassifierCV):
    """
    A classifier that uses random dilated shapelets.

    Parameters
    ----------
    n_shapelets : int, optional
        The number of dilated shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    normalize_prob : float, optional
        The probability of standardizing a shapelet with zero mean and unit
        standard deviation.
    min_shapelet_size : float, optional
        The minimum shapelet size. If None, use the discrete sizes
        in `shapelet_size`.
    max_shapelet_size : float, optional
        The maximum shapelet size. If None, use the discrete sizes
        in `shapelet_size`.
    shapelet_size : array-like, optional
        The size of shapelets.
    lower : float, optional
        The lower percentile to draw distance thresholds above.
    upper : float, optional
        The upper percentile to draw distance thresholds below.
    alphas : array-like of shape (n_alphas,), optional
        Array of alpha values to try.
    fit_intercept : bool, optional
        Whether to calculate the intercept for this model.
    scoring : str, callable, optional
        A string or a scorer callable object with signature
        `scorer(estimator, X, y)`.
    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
    class_weight : dict or 'balanced', optional
        Weights associated with classes in the form `{class_label: weight}`.
    normalize : bool, optional
        Standardize before fitting.
    random_state : int or RandomState, optional
        Controls the random sampling of kernels.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    n_jobs : int, optional
        The number of parallel jobs.

    References
    ----------
    Antoine Guillaume, Christel Vrain, Elloumi Wael
        Random Dilated Shapelet Transform: A New Approach for Time Series Shapelets
        Pattern Recognition and Artificial Intelligence, 2022
    """

    def __init__(  # noqa: PLR0913
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        normalize_prob=0.8,
        min_shapelet_size=None,
        max_shapelet_size=None,
        shapelet_size=None,
        lower=0.05,
        upper=0.1,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        class_weight=None,
        normalize=True,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.shapelet_size = shapelet_size
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.normalize_prob = normalize_prob
        self.lower = lower
        self.upper = upper

    def _get_transform(self, random_state):
        return DilatedShapeletTransform(
            n_shapelets=self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            shapelet_size=self.shapelet_size,
            normalize_prob=self.normalize_prob,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            lower=self.lower,
            upper=self.upper,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


class CastorClassifier(TransformRidgeClassifierCV):
    """
    A dictionary based method using dilated competing shapelets.

    Parameters
    ----------
    n_groups : int, optional
        The number of groups of dilated shapelets.
    n_shapelets : int, optional
        The number of dilated shapelets per group.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    normalize_prob : float, optional
        The probability of standardizing a shapelet with zero mean and unit
        standard deviation.
    shapelet_size : int, optional
        The length of the dilated shapelet.
    lower : float, optional
        The lower percentile to draw distance thresholds above.
    upper : float, optional
        The upper percentile to draw distance thresholds below.
    order : int or array-like, optional
        The order of difference.

        If int, half the groups with corresponding shapelets will be convolved
        with the `order` discrete difference along the time dimension.
    soft_min : bool, optional
        If `True`, use the sum of minimal distances. Otherwise, use the count
        of minimal distances.
    soft_max : bool, optional
        If `True`, use the sum of maximal distances. Otherwise, use the count
        of maximal distances.
    soft_threshold : bool, optional
        If `True`, count the time steps below the threshold for all shapelets.
        Otherwise, count the time steps below the threshold for the shapelet
        with the minimal distance.
    alphas : array-like of shape (n_alphas,), optional
        Array of alpha values to try.
    fit_intercept : bool, optional
        Whether to calculate the intercept for this model.
    scoring : str, callable, optional
        A string or a scorer callable object with signature
        `scorer(estimator, X, y)`.
    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
    class_weight : dict or 'balanced', optional
        Weights associated with classes in the form `{class_label: weight}`.
    normalize : "sparse" or bool, optional
        Standardize before fitting. By default use
        :class:`datasets.preprocess.SparseScaler` to standardize the attributes. Set
        to `False` to disable or `True` to use `StandardScaler`.
    random_state : int or RandomState, optional
        Controls the random sampling of kernels.

        - If `int`, `random_state` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    n_jobs : int, optional
        The number of parallel jobs.
    """

    def __init__(  # noqa: PLR0913
        self,
        n_groups=64,
        n_shapelets=8,
        *,
        metric="euclidean",
        metric_params=None,
        normalize_prob=0.8,
        shapelet_size=11,
        lower=0.05,
        upper=0.1,
        order=1,
        soft_min=True,
        soft_max=False,
        soft_threshold=True,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        class_weight=None,
        normalize="sparse",
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            normalize=normalize,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_groups = n_groups
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.normalize_prob = normalize_prob
        self.shapelet_size = shapelet_size
        self.lower = lower
        self.upper = upper
        self.soft_min = soft_min
        self.soft_max = soft_max
        self.soft_threshold = soft_threshold
        self.order = order

    def _build_pipeline(self):
        pipeline = super()._build_pipeline()
        if self.normalize == "sparse":
            pipeline[1] = ("normalize", SparseScaler())

        return pipeline

    def _get_transform(self, random_state):  # noqa: PLR0912
        random_state = check_random_state(random_state)
        params = dict(
            n_shapelets=self.n_shapelets,
            metric=self.metric,
            metric_params=self.metric_params,
            normalize_prob=self.normalize_prob,
            lower=self.lower,
            upper=self.upper,
            soft_min=self.soft_min,
            soft_max=self.soft_max,
            soft_threshold=self.soft_threshold,
            n_jobs=self.n_jobs,
        )
        n_jobs = effective_n_jobs(self.n_jobs)

        if (
            isinstance(self.order, numbers.Integral)
            and self.order > 0
            and self.n_groups > 1
        ):
            n_groups = self.n_groups // 2
            union_n_jobs = n_jobs // n_groups
            if union_n_jobs < 1:
                union_n_jobs = None

            return make_union(
                CastorTransform(
                    n_groups=n_groups,
                    shapelet_size=self.shapelet_size,
                    random_state=random_state.randint(np.iinfo(np.int32).max),
                    **params,
                ),
                make_pipeline(
                    DiffTransform(order=self.order),
                    CastorTransform(
                        n_groups=n_groups,
                        shapelet_size=self.shapelet_size,
                        random_state=random_state.randint(np.iinfo(np.int32).max),
                        **params,
                    ),
                ),
            )
        else:
            return CastorTransform(
                n_groups=self.n_groups,
                shapelet_size=self.shapelet_size,
                random_state=random_state,
                **params,
            )



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
from wildboar.datasets._filter import make_str_filter


class TestStrFilter:
    def test_n_samples(self):
        f = make_str_filter("n_samples<100")
        assert f("hello", np.zeros((101, 10)), None) is False

    def test_dataset(self):
        f = make_str_filter("dataset=~TwoLead")
        assert f("TwoLeadECG", None, None) is True



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

cdef extern from "pocketfft.h":
    ctypedef struct rfft_plan:
       pass

    ctypedef struct cfft_plan:
       pass

    cdef rfft_plan make_rfft_plan (size_t length) nogil
    cdef void destroy_rfft_plan (rfft_plan plan) nogil
    cdef int rfft_backward(rfft_plan plan, double *c, double fct) nogil
    cdef int rfft_forward(rfft_plan plan, double *c, double fct) nogil
    cdef size_t rfft_length(rfft_plan plan) nogil

    cdef cfft_plan make_cfft_plan(size_t length) nogil
    cdef void destroy_cfft_plan(cfft_plan plan) nogil
    cdef int cfft_backward(cfft_plan plan, double *c, double fct) nogil
    cdef int cfft_forward(cfft_plan plan, double *c, double fct) nogil
    cdef size_t cfft_length(cfft_plan plan) nogil

cdef void fft(complex *x, Py_ssize_t n, double fct) noexcept nogil:
    cdef cfft_plan fft_plan = make_cfft_plan(n)
    cfft_forward(fft_plan, <double *> x, fct)
    destroy_cfft_plan(fft_plan)

cdef void ifft(complex *x, Py_ssize_t n, double fct) noexcept nogil:
    cdef cfft_plan fft_plan = make_cfft_plan(n)
    cfft_backward(fft_plan, <double *> x, fct)
    destroy_cfft_plan(fft_plan)

cdef void rfft(double *x, Py_ssize_t n, double fct) noexcept nogil:
    cdef rfft_plan fft_plan = make_rfft_plan(n)
    rfft_forward(fft_plan, x, fct)
    destroy_rfft_plan(fft_plan)

cdef void irfft(double *x, Py_ssize_t n, double fct) noexcept nogil:
    cdef rfft_plan fft_plan = make_rfft_plan(n)
    rfft_backward(fft_plan, x, fct)
    destroy_rfft_plan(fft_plan)



# Authors: Isak Samsten
# License: BSD 3 clause
import pytest
from sklearn.utils.estimator_checks import check_estimators_pickle
from wildboar.transform import (
    PAA,
    SAX,
    CastorTransform,
    DilatedShapeletTransform,
    HydraTransform,
    IntervalTransform,
    MatrixProfileTransform,
    PivotTransform,
    RandomShapeletTransform,
    RocketTransform,
)
from wildboar.utils._testing import (
    assert_exhaustive_parameter_checks,
    assert_parameter_checks,
)
from wildboar.utils.estimator_checks import check_estimator


@pytest.mark.parametrize(
    "estimator, skip",
    [
        (IntervalTransform(), []),
        (RandomShapeletTransform(), []),
        (RocketTransform(), ["max_size"]),
        (MatrixProfileTransform(), []),
        (PivotTransform(), ["metric_factories"]),
        (SAX(), []),
        (PAA(), []),
        (HydraTransform(), []),
        (DilatedShapeletTransform(), []),
        (CastorTransform(), []),
    ],
)
def test_estimator_checks(estimator, skip):
    check_estimator(estimator)
    assert_exhaustive_parameter_checks(estimator)
    assert_parameter_checks(estimator, skip=skip)



from light_dark import if_not_exists, yield_and_save_plot


@if_not_exists("getting-started", "interval.svg")
def gen_interval_explain(path, force=False):
    from wildboar.datasets import load_two_lead_ecg
    from wildboar.ensemble import ShapeletForestClassifier
    from wildboar.explain import IntervalImportance

    x, y = load_two_lead_ecg()

    clf = ShapeletForestClassifier(n_jobs=-1, n_shapelets=10, random_state=1)
    clf.fit(x, y)

    ex = IntervalImportance(window=8, random_state=1)
    ex.fit(clf, x, y)

    for fig, ax in yield_and_save_plot(path):
        _, mappable = ex.plot(x, y, ax=ax)
        fig.colorbar(mappable, ax=ax)
